import xtrack as xt

import numpy as np
from scipy.constants import c as clight
from scipy.constants import hbar
from scipy.constants import epsilon_0

from cpymad.madx import Madx
import xtrack as xt
import xpart as xp
import xobjects as xo

mad = Madx()
mad.call('../../test_data/sps_thick/sps.seq')

mad.input('beam, particle=proton, pc=26;')
v_mv = 7
mad.call('../../test_data/sps_thick/lhc_q20.str')
mad.use(sequence='sps')
mad.input('twiss, table=tw4d;')
twm4d = mad.table.tw4d

mad.sequence.sps.elements['actcse.31632'].volt = v_mv
mad.sequence.sps.elements['actcse.31632'].freq = 200
mad.sequence.sps.elements['actcse.31632'].lag = 0.5

line = xt.Line.from_madx_sequence(mad.sequence.sps, allow_thick=True,
                                  deferred_expressions=True)
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV,
                                    q0=1, gamma0=mad.sequence.sps.beam.gamma)

# # Optionally thin
# Strategy = xt.slicing.Strategy
# Teapot = xt.slicing.Teapot
# line.slice_thick_elements(slicing_strategies=[
#     Strategy(slicing=Teapot(1)),  # Default
#     Strategy(slicing=Teapot(2), element_type=xt.Bend),
#     Strategy(slicing=Teapot(8), element_type=xt.Quadrupole),
# ])
# line.build_tracker()

tw = line.twiss()

opt = line.match(
    solve=False,
    vary=[
        xt.VaryList(['kqf', 'kqd'], step=1e-5, tag='tune'),
        xt.VaryList(['klsda', 'klsdb', 'klsfa', 'klsfb', 'klsfc'], step=1e-4, tag='chrom'),
    ],
    targets=[
        xt.TargetSet(qx=20.13, qy=20.251, tol=1e-7, tag='tune'),
        xt.TargetSet(dqx=2.0, dqy=2.0, tol=1, tag='chrom'),
        ],
)
opt.solve()

r0 = np.linspace(0, 3, 50)
p = line.build_particles(
    y_norm=r0*np.cos(np.pi/20.),
    py_norm=r0*np.sin(np.pi/20.),
    nemitt_x=1e-6, nemitt_y=1e-6)


line['lod.60702'].knl[3] = 5
line['lod.60702'].knl[2] = 10e-1

line.track(p, num_turns=1000, turn_by_turn_monitor=True)
mon = line.record_last_track

# Compute
# p_co_guess = line.build_particles(y=2e-3)
tw_mt = line.twiss(co_guess={'y': 2e-3}, num_turns=4)

# Inspect and plot
tw_start_turns = tw_mt.rows['_turn_.*']
tw_start_turns.show()
import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(mon.y.flatten(), mon.py.flatten(), '.', markersize=1)
plt.plot(tw_start_turns.y, tw_start_turns.py, '*r')
plt.show()



import xtrack as xt
import xobjects as xo
import numpy as np

line = xt.Line(
    elements=[xt.Drift(length=1.0) for i in range(10)]
)

s_insert = np.array([2.5, 5.5, 7])
l_insert = np.array([1.0, 1.0, 1.0])
ele_insert = [xt.Sextupole(length=l) for l in l_insert]

line._insert_thick_elements_at_s(
    element_names=[f'insertion_{i}' for i in range(len(s_insert))],
    elements=ele_insert,
    at_s=s_insert
)

tt = line.get_table()

assert np.all(tt.name == ['e0', 'e1', 'e2..0', 'insertion_0', 'e3..1', 'e4', 'e5..0',
       'insertion_1', 'e6..1', 'insertion_2', 'e8', 'e9', '_end_point'])
xo.assert_allclose(tt.s, [ 0. ,  1. ,  2. ,  2.5,  3.5,  4. ,  5. ,
                                 5.5,  6.5,  7. ,  8. , 9. , 10. ])

assert np.all(tt.element_type == ['Drift', 'Drift', 'DriftSlice', 'Sextupole', 'DriftSlice', 'Drift',
       'DriftSlice', 'Sextupole', 'DriftSlice', 'Sextupole', 'Drift',
       'Drift', ''])


import time
import numpy as np
import xtrack as xt

line = xt.Line.from_json('lhc_thin.json')

tt = line.get_table()

# Identify active elements
tt_active = tt.rows[
    (tt.element_type != 'Drift') & (tt.element_type != 'Marker')]

# Generate an aperture and a marker to be put in front of  each active element
insertions = []
for ii, (nn, ss) in enumerate(zip(tt_active.name, tt_active.s)):
    print(f'{ii}/{len(tt_active.name)}', end='\r', flush=True)

    aa = xt.LimitRect(min_x=-0.1, max_x=0.1, min_y=-0.1, max_y=0.1)
    mm = xt.Marker()
    insertions.append((ss, [(f'aper_{nn}', aa), (f'marker_{nn}', mm)]))

# Generate an aperture and a marker to be added every 100 m
for ss in np.arange(100, tt.s[-1], 100):
    aa = xt.LimitRect(min_x=-0.2, max_x=0.2, min_y=-0.2, max_y=0.2)
    mm = xt.Marker()
    insertions.append((ss, [(f'aper_{ss}', aa), (f'marker_{ss}', mm)]))

line.discard_tracker()

# Insert all
line._insert_thin_elements_at_s(insertions)

# Inspect:
tt_after = line.get_table()
tt_after.rows['marker_20000.0<<30': 'marker_20000.0>>30'].show()



import numpy as np

import xtrack as xt

b = xt.Bend(k0=0.2, h=0.1, length=1.0)

line = xt.Line(elements=[b])
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, beta0=0.5)
line.reset_s_at_end_turn = False
line.build_tracker()

p0 = line.build_particles(x=0.01, px=0.02, y=0.03, py=0.04,
                         zeta=0.05, delta=0.01)

p1 = p0.copy()
line.track(p1)

p2 = p1.copy()
line.track(p2, backtrack=True)

print('\nBend:')
for nn in 's x px y py zeta delta'.split():
    print(f"{nn}: {getattr(p0, nn)[0]:.6e} {getattr(p1, nn)[0]:.6e} {getattr(p2, nn)[0]:.6e}")

assert np.allclose(p2.s, p0.s, atol=1e-15, rtol=0)
assert np.allclose(p2.x, p0.x, atol=1e-15, rtol=0)
assert np.allclose(p2.px, p0.px, atol=1e-15, rtol=0)
assert np.allclose(p2.y, p0.y, atol=1e-15, rtol=0)
assert np.allclose(p2.py, p0.py, atol=1e-15, rtol=0)
assert np.allclose(p2.zeta, p0.zeta, atol=1e-15, rtol=0)
assert np.allclose(p2.delta, p0.delta, atol=1e-15, rtol=0)

p3 = p1.copy()
line.configure_bend_model(core='full')
line.track(p3, backtrack=True)

assert np.all(p3.state == -30)

p4 = p1.copy()
line.configure_bend_model(core='expanded')
b.num_multipole_kicks = 3
line.track(p4, backtrack=True)

assert np.all(p4.state == -31)

# Same for quadrupole
q = xt.Quadrupole(k1=0.2, length=1.0)
assert np.allclose(p2.s, p0.s, atol=1e-15, rtol=0)
assert np.allclose(p2.x, p0.x, atol=1e-15, rtol=0)
assert np.allclose(p2.px, p0.px, atol=1e-15, rtol=0)
assert np.allclose(p2.y, p0.y, atol=1e-15, rtol=0)
assert np.allclose(p2.py, p0.py, atol=1e-15, rtol=0)
assert np.allclose(p2.zeta, p0.zeta, atol=1e-15, rtol=0)
assert np.allclose(p2.delta, p0.delta, atol=1e-15, rtol=0)

line = xt.Line(elements=[q])
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, beta0=0.5)
line.reset_s_at_end_turn = False
line.build_tracker()

p0 = line.build_particles(x=0.01, px=0.02, y=0.03, py=0.04,
                            zeta=0.05, delta=0.01)

p1 = p0.copy()
line.track(p1)

p2 = p1.copy()
line.track(p2, backtrack=True)

print('\nQuadrupole:')
for nn in 's x px y py zeta delta'.split():
    print(f"{nn}: {getattr(p0, nn)[0]:.6e} {getattr(p1, nn)[0]:.6e} {getattr(p2, nn)[0]:.6e}")

p4 = p1.copy()
q.num_multipole_kicks = 4
line.track(p4, backtrack=True)

assert np.all(p4.state == -31)

de = xt.DipoleEdge(e1=0.1, k=3, fint=0.3)
line = xt.Line(elements=[de])
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, beta0=0.5)
line.reset_s_at_end_turn = False
line.build_tracker()

p0 = line.build_particles(x=0.01, px=0.02, y=0.03, py=0.04,
                            zeta=0.05, delta=0.01)

p1 = p0.copy()
line.track(p1)

assert np.all(p1.state == 1)

line.configure_bend_model(edge='full')
p2 = p1.copy()
line.track(p2, backtrack=True)

assert np.all(p2.state == -32)


import xtrack as xt
import xobjects as xo
import numpy as np

def test_periodic_symmetric_twiss_and_match():

    # Build line with half a cell
    half_cell = xt.Line(
        elements={
            'start_cell': xt.Marker(),
            'drift0': xt.Drift(length=1.),
            'qf1': xt.Quadrupole(k1=0.027/2, length=1.),
            'drift1_1': xt.Drift(length=1),
            'bend1': xt.Bend(k0=3e-4, h=3e-4, length=45.),
            'drift1_2': xt.Drift(length=1.),
            'qd1': xt.Quadrupole(k1=-0.0271/2, length=1.),
            'drift2': xt.Drift(length=1),
            'mid_cell': xt.Marker(),
        }
    )
    half_cell.particle_ref = xt.Particles(p0c=2e9)

    tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', # <--- periodic-symmetric boundary
                                    strengths=True # to get the strengths in table
                                    )

    cell = xt.Line(
        elements={
            'start_cell': xt.Marker(),
            'drift0': xt.Drift(length=1.),
            'qf1':    xt.Quadrupole(k1=0.027/2, length=1.),
            'drift1': xt.Drift(length=1),
            'bend1':  xt.Bend(k0=3e-4, h=3e-4, length=45.),
            'drift2': xt.Drift(length=1.),
            'qd1':    xt.Quadrupole(k1=-0.0271/2, length=1.),
            'drift3': xt.Drift(length=1),
            'mid_cell': xt.Marker(),
            'drift4': xt.Replica('drift3'),
            'qd2':    xt.Replica('qd1'),
            'drift5': xt.Replica('drift2'),
            'bend2':  xt.Replica('bend1'),
            'drift6': xt.Replica('drift1'),
            'qf2':    xt.Replica('qf1'),
            'drift7': xt.Replica('drift0'),
            'end_cell': xt.Marker(),
        }
    )
    cell.particle_ref = xt.Particles(p0c=2e9)
    tw_cell = cell.twiss4d(strengths=True)

    xo.assert_allclose(tw_half_cell.betx[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].betx, atol=0, rtol=1e-8)
    xo.assert_allclose(tw_half_cell.bety[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].bety, atol=0, rtol=1e-8)
    xo.assert_allclose(tw_half_cell.alfx[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].alfx, atol=1e-8, rtol=0)
    xo.assert_allclose(tw_half_cell.alfy[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].alfy, atol=1e-8, rtol=0)
    xo.assert_allclose(tw_half_cell.dx[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].dx, atol=1e-8, rtol=0)
    xo.assert_allclose(tw_half_cell.dpx[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].dpx, atol=1e-8, rtol=0)

    xo.assert_allclose(tw_half_cell.ax_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].ax_chrom, atol=1e-5, rtol=0)
    xo.assert_allclose(tw_half_cell.ay_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].ay_chrom, atol=1e-5, rtol=0)
    xo.assert_allclose(tw_half_cell.bx_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].bx_chrom, atol=1e-5, rtol=0)
    xo.assert_allclose(tw_half_cell.by_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].by_chrom, atol=1e-5, rtol=0)

    xo.assert_allclose(tw_half_cell.qx, tw_cell.qx / 2, atol=1e-9, rtol=0)
    xo.assert_allclose(tw_half_cell.qy, tw_cell.qy / 2, atol=1e-9, rtol=0)
    xo.assert_allclose(tw_half_cell.dqx, tw_cell.dqx / 2, atol=1e-6, rtol=0)
    xo.assert_allclose(tw_half_cell.dqy, tw_cell.dqy / 2, atol=1e-6, rtol=0)

    for ll in [cell, half_cell]:
        ll.vars['kqf'] = 0.027/2
        ll.vars['kqd'] = -0.0271/2
        ll.element_refs['qf1'].k1 = ll.vars['kqf']
        ll.element_refs['qd1'].k1 = ll.vars['kqd']

    tw_cell = cell.twiss4d(strengths=True)
    tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', strengths=True)

    xo.assert_allclose(tw_cell.mux[-1], 0.231, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_cell.muy[-1], 0.233, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_half_cell.mux[-1], 0.231 / 2, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_half_cell.muy[-1], 0.233 / 2, atol=1e-3, rtol=0)

    opt_halfcell = half_cell.match(
        method='4d',
        start='start_cell', end='mid_cell',
        init='periodic_symmetric',
        targets=xt.TargetSet(mux=0.2501/2, muy=0.2502/2, at='mid_cell'),
        vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
    )

    tw_cell = cell.twiss4d(strengths=True)
    tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', strengths=True)

    xo.assert_allclose(tw_half_cell.mux[-1], 0.2501 / 2, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_half_cell.muy[-1], 0.2502 / 2, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_cell.mux[-1], 0.231, atol=1e-3, rtol=0) # unaffected
    xo.assert_allclose(tw_cell.muy[-1], 0.233, atol=1e-3, rtol=0) # unaffected

    opt_cell = cell.match(
        method='4d',
        start='start_cell', end='end_cell',
        init='periodic',
        targets=xt.TargetSet(mux=0.2501, muy=0.2502, at='end_cell'),
        vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
    )

    xo.assert_allclose(half_cell.vv['kqf'], cell.vv['kqf'], rtol=1e-9, atol=0)
    xo.assert_allclose(half_cell.vv['kqd'], cell.vv['kqd'], rtol=1e-9, atol=0)

    tw_cell = cell.twiss4d(strengths=True)
    tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', strengths=True)

    xo.assert_allclose(tw_half_cell.mux[-1], 0.2501 / 2, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_half_cell.muy[-1], 0.2502 / 2, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_cell.mux[-1], 0.2501, atol=1e-3, rtol=0)
    xo.assert_allclose(tw_cell.muy[-1], 0.2502, atol=1e-3, rtol=0)

    xo.assert_allclose(tw_half_cell.betx[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].betx, atol=0, rtol=1e-8)
    xo.assert_allclose(tw_half_cell.bety[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].bety, atol=0, rtol=1e-8)
    xo.assert_allclose(tw_half_cell.alfx[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].alfx, atol=1e-8, rtol=0)
    xo.assert_allclose(tw_half_cell.alfy[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].alfy, atol=1e-8, rtol=0)
    xo.assert_allclose(tw_half_cell.dx[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].dx, atol=1e-8, rtol=0)
    xo.assert_allclose(tw_half_cell.dpx[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].dpx, atol=1e-8, rtol=0)
    xo.assert_allclose(tw_half_cell.ddx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].ddx, atol=1e-7, rtol=1e-8)

    xo.assert_allclose(tw_half_cell.ax_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].ax_chrom, atol=1e-5, rtol=0)
    xo.assert_allclose(tw_half_cell.ay_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].ay_chrom, atol=1e-5, rtol=0)
    xo.assert_allclose(tw_half_cell.bx_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].bx_chrom, atol=1e-5, rtol=0)
    xo.assert_allclose(tw_half_cell.by_chrom[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].by_chrom, atol=1e-5, rtol=0)
    xo.assert_allclose(tw_half_cell.ddx[:-1], # remove '_end_point'
                        tw_cell.rows[:'mid_cell'].ddx, atol=1e-7, rtol=1e-8)

    xo.assert_allclose(tw_half_cell.qx, tw_cell.qx / 2, atol=1e-9, rtol=0)
    xo.assert_allclose(tw_half_cell.qy, tw_cell.qy / 2, atol=1e-9, rtol=0)
    xo.assert_allclose(tw_half_cell.dqx, tw_cell.dqx / 2, atol=1e-6, rtol=0)
    xo.assert_allclose(tw_half_cell.dqy, tw_cell.dqy / 2, atol=1e-6, rtol=0)

    tw_off_mom_cell = cell.twiss4d(strengths=True, delta0=1e-3)
    tw_off_mom_half_cell = half_cell.twiss4d(
        init='periodic_symmetric', strengths=True, delta0=1e-3)

    xo.assert_allclose(tw_off_mom_half_cell.x[:-1],
                    tw_off_mom_cell.rows[:'mid_cell'].x, atol=1e-12, rtol=0)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib
import pickle
import math

import numpy as np
import pytest

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts
from xtrack import Line, Node, Multipole

test_data_folder = pathlib.Path(
            __file__).parent.joinpath('../test_data').absolute()


def test_simplification_methods():

    line = xt.Line(
        elements=([xt.Drift(length=0)] # Start line marker
                    + [xt.Drift(length=1) for _ in range(5)]
                    + [xt.Drift(length=0)] # End line marker
            )
        )

    # Test merging of drifts
    line.insert_element(element=xt.Cavity(), name='cav', at_s=3.3)
    assert isinstance(line['e4..0'], xt.DriftSlice)
    line._replace_with_equivalent_elements()
    assert isinstance(line['e4..0'], xt.Drift)
    line.merge_consecutive_drifts(inplace=True)
    assert len(line.element_names) == 3
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5
    xo.assert_allclose(line[line.element_names[0]].length, 3.3, rtol=0, atol=1e-12)
    assert isinstance(line[line.element_names[1]], xt.Cavity)
    xo.assert_allclose(line[line.element_names[2]].length, 1.7, rtol=0, atol=1e-12)

    # Test merging of drifts, while keeping one
    line.insert_element(element=xt.Drift(length=1), name='drift1', at_s=1.2)
    line.insert_element(element=xt.Drift(length=1), name='drift2', at_s=2.2)
    line._replace_with_equivalent_elements()
    line.merge_consecutive_drifts(inplace=True, keep=['drift2'])
    assert len(line.element_names) == 5
    assert 'drift2' in line.element_names
    assert 'drift1' not in line.element_names
    line._replace_with_equivalent_elements()
    line.merge_consecutive_drifts(inplace=True)

    # Test removing of zero-length drifts
    line.insert_element(element=xt.Drift(length=0), name='dzero1', at_s=3.3)
    line.insert_element(element=xt.Drift(length=0), name='dzero2', at_s=3.3)
    assert len(line.element_names) == 5
    line._replace_with_equivalent_elements()
    line.remove_zero_length_drifts(inplace=True, keep='dzero2')
    assert len(line.element_names) == 4
    assert 'dzero2' in line.element_names
    assert 'dzero1' not in line.element_names
    line._replace_with_equivalent_elements()
    line.remove_zero_length_drifts(inplace=True)

    # Test merging of multipoles
    line._var_management = None
    line.insert_element(element=xt.Multipole(knl=[1, 0, 3], ksl=[0, 20, 0]), name='m1', at_s=3.3)
    line.insert_element(element=xt.Multipole(knl=[4, 2], ksl=[10, 40]), name='m2', at_s=3.3)
    line.insert_element(element=xt.Multipole(knl=[0, 3, 8], ksl=[2, 0, 17]), name='m3', at_s=3.3)
    line.insert_element(element=xt.Multipole(knl=[2, 0, 0], ksl=[40]), name='m4', at_s=3.3)
    assert len(line.element_names) == 7
    line._replace_with_equivalent_elements()
    line.merge_consecutive_multipoles(inplace=True, keep='m3')
    assert len(line.element_names) == 6
    assert 'm3' in line.element_names
    # We merged the first two multipoles
    joined_mult = [ name for name in line.element_names if 'm1' in name]
    assert len(joined_mult) == 1
    joined_mult = joined_mult[0]
    assert 'm2' in joined_mult
    xo.assert_allclose(line[joined_mult].knl, [5,2,3], rtol=0, atol=1e-15)
    xo.assert_allclose(line[joined_mult].ksl, [10,60,0], rtol=0, atol=1e-15)
    # Merging all
    line._replace_with_equivalent_elements()
    line.merge_consecutive_multipoles(inplace=True)
    assert len(line.element_names) == 4
    xo.assert_allclose(line[line.element_names[1]].knl, [7,5,11], rtol=0, atol=1e-15)
    xo.assert_allclose(line[line.element_names[1]].ksl, [52,60,17], rtol=0, atol=1e-15)

    # Test removing inactive multipoles
    line.insert_element(element=xt.Multipole(knl=[0, 8, 1], ksl=[0, 20, 30]), name='m5', at_s=3.3)
    line.insert_element(element=xt.Multipole(knl=[2, 0, 3], ksl=[10, 34, 15]), name='m6', at_s=3.3)
    line.remove_inactive_multipoles(inplace=True)
    assert len(line.element_names) == 6
    line['m5'].knl[:] = 0
    line['m5'].ksl[:] = 0
    line['m6'].knl[:] = 0
    line['m6'].ksl[:] = 0
    line._replace_with_equivalent_elements()
    line.remove_inactive_multipoles(inplace=True, keep='m5')
    assert len(line.element_names) == 5
    assert 'm5' in line.element_names
    assert 'm6' not in line.element_names

    # Test removing markers
    line.insert_element(element=xt.Marker(), name='marker1', at_s=3.3)
    line.insert_element(element=xt.Marker(), name='marker2', at_s=3.3)
    assert 'marker1' in line.element_names
    assert 'marker2' in line.element_names
    line._replace_with_equivalent_elements()
    line.remove_markers(keep='marker2')
    assert 'marker1' not in line.element_names
    assert 'marker2' in line.element_names
    line.insert_element(element=xt.Marker(), name='marker4', at_s=3.3)
    line.insert_element(element=xt.Marker(), name='marker3', at_s=3.3)
    assert 'marker2' in line.element_names
    assert 'marker3' in line.element_names
    assert 'marker4' in line.element_names
    line._replace_with_equivalent_elements()
    line.remove_markers()
    assert 'marker2' not in line.element_names
    assert 'marker3' not in line.element_names
    assert 'marker4' not in line.element_names


def test_remove_redundant_apertures():

    # Lattice:
    # D1-A1-M-D2 D3-A2-M-D4 D5-A3-M-D6 D7-A4-M-D8 D9-A5-M-D10
    elements = []
    for _ in range(5):
        elements += [
            xt.Drift(length=0.6),
            xt.LimitRect(min_x=-0.3, max_x=0.3,min_y=-0.3, max_y=0.3),
            xt.Marker(),
            xt.Drift(length=0.4)
        ]
    line = xt.Line(elements=elements)
    original_line = line.copy()

    # Test removing all consecutive middle apertures
    assert len(line.element_names) == 20
    all_aper = [nn for nn in line.element_names if xt._is_aperture(line[nn], line)]
    all_aper_pos = [line.get_s_position(ap) for ap in all_aper]
    line.remove_redundant_apertures()
    line.remove_markers()
    line.merge_consecutive_drifts()
    # The lattice is now D1-A1-DD-A5-D10
    assert len(line.element_names) == 5
    # Verify that only the first and last aperture are kept
    new_aper = [nn for nn in line.element_names if xt._is_aperture(line[nn], line)]
    assert new_aper == [all_aper[0], all_aper[-1]]
    new_aper_pos = [line.get_s_position(ap) for ap in new_aper]
    assert new_aper_pos == [all_aper_pos[0], all_aper_pos[-1]]

    # Test removing all consecutive middle apertures, but
    # keep the 4th one (and hence also the 5th)
    line = original_line.copy()
    assert len(line.element_names) == 20
    all_aper = [nn for nn in line.element_names if xt._is_aperture(line[nn], line)]
    all_aper_pos = [line.get_s_position(ap) for ap in all_aper]
    line.remove_redundant_apertures(keep=all_aper[3])
    line.remove_markers()
    line.merge_consecutive_drifts()
    # The lattice is now D1-A1-DD-A4-DD-A5-D10
    assert len(line.element_names) == 7
    # Verify that only the first, fourth, and last aperture are kept
    new_aper = [nn for nn in line.element_names if xt._is_aperture(line[nn], line)]
    assert new_aper == [all_aper[0], all_aper[3], all_aper[-1]]
    new_aper_pos = [line.get_s_position(ap) for ap in new_aper]
    assert new_aper_pos == [all_aper_pos[0], all_aper_pos[3], all_aper_pos[-1]]

    # Test removing all consecutive middle apertures, but
    # the 9th Drift needs to have an aperture. This should
    # give the same result as above
    line = original_line.copy()
    assert len(line.element_names) == 20
    all_aper = [nn for nn in line.element_names if xt._is_aperture(line[nn], line)]
    all_aper_pos = [line.get_s_position(ap) for ap in all_aper]
    all_drifts = [nn for nn in line.element_names if xt._is_drift(line[nn], line)]
    line.remove_redundant_apertures(drifts_that_need_aperture=all_drifts[8])
    line.remove_markers()
    line.merge_consecutive_drifts()
    # The lattice is now D1-A1-DD-A4-DD-A5-D10
    assert len(line.element_names) == 7
    # Verify that only the first, fourth, and last aperture are kept
    new_aper = [nn for nn in line.element_names if xt._is_aperture(line[nn], line)]
    assert new_aper == [all_aper[0], all_aper[3], all_aper[-1]]
    new_aper_pos = [line.get_s_position(ap) for ap in new_aper]
    assert new_aper_pos == [all_aper_pos[0], all_aper_pos[3], all_aper_pos[-1]]

    # All apertures are different, none should be removed
    elements = []
    for i in range(5):
        elements += [
            xt.Drift(length=0.6),
            xt.LimitRect(min_x=-0.3+i*0.01, max_x=0.3,min_y=-0.3, max_y=0.3),
            xt.Marker(),
            xt.Drift(length=0.4)
        ]
    line = xt.Line(elements=elements)
    original_line = line.copy()
    line.remove_redundant_apertures()
    assert xt._lines_equal(line, original_line)

def test_redundant_apertures():
    sequence = [
        ('a1', xt.LimitRect(min_x=-0.3, max_x=0.3, min_y=-0.3, max_y=0.3)),
        ('d0', xt.Drift(length=0.6)),
        ('a1', xt.LimitRect(min_x=-0.3, max_x=0.3, min_y=-0.3, max_y=0.3)),
        ('d1', xt.Drift(length=0.4)),
        ('a2', xt.Replica('a1')),
        ('m2', xt.Marker()),
        ('d2..1', xt.Drift(length=0.2)),
        ('d2..2', xt.Drift(length=0.2)),
        ('a3', xt.LimitRect(min_x=-0.3, max_x=0.3, min_y=-0.3, max_y=0.3)),
        ('d3', xt.Drift(length=0.4)),
    ]
    line = xt.Line(elements=dict(sequence), element_names=[n for n, _ in sequence])

    line.remove_redundant_apertures()

    expected_names = ['d0', 'a1', 'd1', 'm2', 'd2..1', 'd2..2', 'a3', 'd3']
    assert line.element_names == expected_names

def test_insert():

    line0 = xt.Line(
        elements = [xt.Drift(length=1) for _ in range(5)]
    )

    line = line0.copy()
    assert np.all(np.array([0,1,2,3,4]) == np.array(line.get_s_elements()))
    assert np.all(np.array([0,1,2,3,4]) == np.array(line.get_s_elements(mode='upstream')))
    assert np.all(np.array([1,2,3,4,5]) == np.array(line.get_s_elements(mode='downstream')))

    assert line.get_s_position(at_elements='e3') == 3.
    assert np.isscalar(line.get_s_position(at_elements='e3'))
    assert len(line.get_s_position(at_elements=['e3'])) == 1
    assert np.all(np.array([4,2]) == np.array(line.get_s_position(at_elements=['e4', 'e2'])))

    line.insert_element(element=xt.Cavity(), name="cav", at_s=3.3)
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5
    assert line.get_s_position('cav') == 3.3
    assert len(line.elements) == 7

    line = line0.copy()
    line.insert_element(element=xt.Drift(length=0.2), at_s=0.11, name='inserted_drift')
    assert line.get_s_position('inserted_drift') == 0.11
    assert len(line.elements) == 7
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
                ['e0..0', 'inserted_drift', 'e0..2', 'e1', 'e2', 'e3', 'e4']))])
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5

    line = line0.copy()
    line.insert_element(element=xt.Drift(length=0.2), at_s=0.95, name='inserted_drift')
    assert line.get_s_position('inserted_drift') == 0.95
    assert len(line.elements) == 6
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
                ['e0..0', 'inserted_drift', 'e1..1', 'e2', 'e3', 'e4']))])
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5

    line = line0.copy()
    line.insert_element(element=xt.Drift(length=0.2), at_s=1.0, name='inserted_drift')
    assert line.get_s_position('inserted_drift') == 1.
    assert len(line.elements) == 6
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
                ['e0', 'inserted_drift', 'e1..1', 'e2', 'e3', 'e4']))])
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5

    line = line0.copy()
    line.insert_element(element=xt.Drift(length=0.2), at_s=0.8, name='inserted_drift')
    assert line.get_s_position('inserted_drift') == 0.8
    assert len(line.elements) == 6
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
                ['e0..0', 'inserted_drift', 'e1', 'e2', 'e3', 'e4']))])
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5

    line = line0.copy()
    line.insert_element(element=xt.LimitEllipse(a=1, b=1), at_s=2.1, name='aper')
    assert line.get_s_position('aper') == 2.1
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
                ['e0', 'e1', 'e2..0', 'aper', 'e2..1', 'e3', 'e4']))])
    line.insert_element(element=xt.Drift(length=0.8), at_s=1.9, name="newdrift")
    assert line.get_s_position('newdrift') == 1.9
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
                ['e0', 'e1..0', 'newdrift', 'e2..1..1', 'e3', 'e4']))])

    # Check preservation of markers
    elements = []
    enames = []

    for ii in range(5):
        elements.append(xt.Drift(length=1))
        enames.append(f'd{ii}')
        elements.append(xt.Drift(length=0))
        enames.append(f'm{ii}')

    line = xt.Line(elements=elements, element_names=enames)
    line.insert_element(element=xt.Drift(length=1.), at_s=1.0, name='inserted_drift')
    assert line.get_s_position('inserted_drift') == 1.
    assert len(line.elements) == 10
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
        ['d0', 'm0', 'inserted_drift', 'm1', 'd2', 'm2', 'd3', 'm3', 'd4', 'm4']))])
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5

    line.insert_element(element=xt.Cavity(), at_s=3.0, name='cav0')
    line.insert_element(element=xt.Cavity(), at_s=3.0, name='cav1')
    assert len(line.elements) == 12
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
        ['d0', 'm0', 'inserted_drift', 'm1', 'd2', 'cav1', 'cav0', 'm2', 'd3',
        'm3', 'd4', 'm4']))])
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5
    assert line.get_s_position('cav0') == 3.
    assert line.get_s_position('cav1') == 3.

    line = xt.Line(elements=elements, element_names=enames)
    line.insert_element(element=xt.Drift(length=0.2), at_s=0.95, name='inserted_drift')
    assert line.get_s_position('inserted_drift') == 0.95
    assert len(line.elements) == 10
    assert np.all([nn==nnref for nn, nnref in list(zip(line.element_names,
                ['d0..0', 'inserted_drift', 'd1..1', 'm1', 'd2', 'm2', 'd3',
                'm3', 'd4', 'm4']))])
    assert line.get_length() == line.get_s_elements(mode='downstream')[-1] == 5


def test_insert_omp():
    ctx = xo.ContextCpu(omp_num_threads='auto')
    buffer = ctx.new_buffer()

    drift = xt.Drift(length=2, _buffer=buffer)
    multipole = xt.Multipole(knl=[1], _buffer=buffer)

    line = xt.Line(elements=[drift], element_names=['dr'])
    line.insert_element(element=multipole, at_s=1, name='mp')
    line.build_tracker()

    assert line._buffer is line['dr..0']._buffer
    assert line['dr..0']._buffer is line['mp']._buffer
    assert line['mp']._buffer is line['dr..1']._buffer
    assert line._context.omp_num_threads == 'auto'


def test_to_pandas():
    line = xt.Line(elements=[
        xt.Drift(length=1), xt.Cavity(), xt.Drift(length=1)])

    df = line.to_pandas()

    assert tuple(df.columns) == (
        's', 'element_type', 'name', 'isthick', 'isreplica', 'parent_name',
       'iscollective', 'element', 's_start', 's_center', 's_end')
    assert len(df) == 4

def test_check_aperture():

    class ThickElement:
        length = 2.
        isthick = True

    line = xt.Line(
        elements={
            'dum': xt.Drift(length=0),
            'dr1': xt.Drift(length=1),
            'm1_ap': xt.LimitEllipse(a=1e-2, b=1e-2),
            'm1': xt.Multipole(knl=[1]),
            'dr2': xt.Drift(length=1),
            'm2': xt.Multipole(knl=[1]),
            'dr3': xt.Drift(length=1),
            'th1_ap_front': xt.LimitEllipse(a=1e-2, b=1e-2),
            'th1': ThickElement(),
            'th1_ap_back': xt.LimitEllipse(a=1e-2, b=1e-2),
            'dr4': xt.Drift(length=1),
            'th2': ThickElement(),
            'th2_ap_back': xt.LimitEllipse(a=1e-2, b=1e-2),
            'dr5': xt.Drift(length=1),
            'th3_ap_front': xt.LimitEllipse(a=1e-2, b=1e-2),
            'th3': ThickElement(),
            'dr6': xt.Drift(length=1),
            'th4_ap_entry': xt.LimitEllipse(a=1e-2, b=1e-2),
            'th4': ThickElement(),
            'th4_ap_exit': xt.LimitEllipse(a=1e-2, b=1e-2),
        },
        element_names=['dr1', 'm1_ap', 'dum', 'm1', 'dr2', 'm2', 'dr3',
                       'th1_ap_front', 'dum', 'th1', 'dum', 'th1_ap_back',
                       'dr4', 'th2', 'th2_ap_back',
                       'dr5', 'th3_ap_front', 'th3', 'dr6',
                       'th4_ap_entry', 'th4', 'th4_ap_exit'])

    df = line.check_aperture()

    expected_miss_upstream = [nn in ('m2', 'th2') for nn in df['name'].values]
    expected_miss_downstream = [nn in ('m1', 'm2', 'th3') for nn in df['name'].values]
    expected_problem_flag = np.array(expected_miss_upstream) | (df.isthick &
                                        np.array(expected_miss_downstream))

    assert np.all(df['misses_aperture_upstream'].values == expected_miss_upstream)
    assert np.all(df['misses_aperture_downstream'].values == expected_miss_downstream)
    assert np.all(df['has_aperture_problem'].values == expected_problem_flag)

def test_to_dict():
    line = xt.Line(
        elements={
            'm': xt.Multipole(knl=[1, 2]),
            'd': xt.Drift(length=1),
        },
        element_names=['m', 'd', 'm', 'd']
    )
    result = line.to_dict()

    assert len(result['elements']) == 2
    assert result['element_names'] == ['m', 'd', 'm', 'd']

    assert result['elements']['m']['__class__'] == 'Multipole'
    assert (result['elements']['m']['knl'] == [1, 2]).all()

    assert result['elements']['d']['__class__'] == 'Drift'
    assert result['elements']['d']['length'] == 1

    assert result['metadata'] == line.metadata

def test_from_dict_legacy():
    test_dict = {
        'elements': [
            {'__class__': 'Multipole', 'knl': [1, 2]},
            {'__class__': 'Drift', 'length': 1},
        ],
        'element_names': ['mn1', 'd1'],
    }
    result = xt.Line.from_dict(test_dict)

    assert len(result.elements) == 2

    assert isinstance(result.elements[0], xt.Multipole)
    assert (result.elements[0].knl == [1, 2]).all()

    assert isinstance(result.elements[1], xt.Drift)
    assert result.elements[1].length == 1

    assert result.element_names == ['mn1', 'd1']

@pytest.mark.parametrize('under_test', ['copy', 'from_to_dict'])
def test_copy_to_dict_from_dict_config(under_test):

    # Step 1: Create a line object
    line = xt.Line()

    # Step 2: Build the tracker
    line.build_tracker()
    line.metadata['hello'] = 'world'

    # Step 3: Configure radiation model
    line.configure_radiation(model='mean')

    # Step 5: Serialize and deserialize the line
    if under_test == 'copy':
        l2 = line.copy()
    elif under_test == 'from_to_dict':
        line_dict = line.to_dict()
        l2 = xt.Line.from_dict(line_dict)
    else:
        raise ValueError(f'Unknown test {under_test}')

    assert np.all(
        np.sort(list(line.config.keys())) == np.sort(list(l2.config.keys())))

    for key in line.config.keys():
        assert line.config[key] == l2.config[key]

    assert line.metadata['hello'] == l2.metadata['hello']


def test_from_dict_current():
    test_dict = {
        'elements': {
            'mn': {
                '__class__': 'Multipole',
                'knl': [1, 2],
            },
            'ms': {
                '__class__': 'Multipole',
                'ksl': [3],
            },
            'd': {
                '__class__': 'Drift',
                'length': 4,
            },
        },
        'element_names': ['mn', 'd', 'ms', 'd'],
        'metadata' : {
            'config_knobs_and_tuning': {
                'knob_settings': {
                    'on_x1': 135.0,
                },
            },
        },
    }
    line = xt.Line.from_dict(test_dict)

    assert line.element_names == ['mn', 'd', 'ms', 'd']
    mn, d1, ms, d2 = line.elements

    assert isinstance(mn, xt.Multipole)
    assert (mn.knl == [1, 2]).all()

    assert isinstance(ms, xt.Multipole)
    assert (ms.ksl == [3]).all()

    assert isinstance(d1, xt.Drift)
    assert d1.length == 4

    assert d2 is d1

    assert line.metadata == test_dict['metadata']


def test_from_sequence():
    # direct element definition
    # -------------------------
    line = Line.from_sequence([
        Node(3, Multipole(knl=[1])),
        Node(7, Multipole(), name='quad1'),
        Node(1, Multipole(), name='bend1', from_='quad1'),
        ], 10)
    assert line.get_length() == 10
    assert len(line.elements) == 7
    for i, l in enumerate([3, 0, 4, 0, 1, 0, 2]):
        cls = xt.Multipole if i%2 else xt.Drift
        assert isinstance(line.elements[i], cls)
        xo.assert_allclose(line.elements[i].length, l)

    # using pre-defined elements by name
    # ----------------------------------
    elements = {
        'quad': Multipole(length=0.3, knl=[0, +0.50]),
        'bend': Multipole(length=0.5, knl=[np.pi / 12], hxl=[np.pi / 12]),
    }
    line = Line.from_sequence(length=10, nodes=[
        Node(1, 'quad'),
        Node(1, 'quad', name='quad3', from_=3),
        Node(2, 'bend', from_='quad3', name='bend2'),
    ], elements=elements, auto_reorder=True)
    assert line.get_length() == 10
    assert len(line.elements) == 7
    for i, l in enumerate([1, 0.3, 3, 0.3, 2, 0.5, 4]):
        cls = xt.Multipole if i%2 else xt.Drift
        assert isinstance(line.elements[i], cls)
        xo.assert_allclose(line.elements[i].length, l)
    assert line.elements[1] == line.elements[3]
    assert line.element_names[1] == 'quad'
    assert line.element_names[3] == 'quad3'
    assert line.element_names[5] == 'bend2'
    assert line.elements[5] == elements['bend']

    # using nested sequences
    # ----------------------
    sequences = {
        'arc': [Node(1, 'quad'), Node(4, 'bend', from_='quad')],
    }
    sext = Multipole(knl=[0, 0, 0.1])
    line = Line.from_sequence([
        Node(0, 'arc', name='section_1'),
        Node(10, 'arc', name='section_2'),
        Node(3, sext, from_='section_2', name='sext'),
        Node(3, [Node(1, 'quad')], name='section_3', from_='sext'),
    ], length=20, elements=elements, sequences=sequences, auto_reorder=True, naming_scheme='{}_{}')

    assert line.get_length() == 20
    assert len(line.elements) == 18

    assert line.get_s_position()[line.element_names.index('section_1')] == 0
    assert isinstance(line.elements[line.element_names.index('section_1')], xt.Marker)
    assert line.get_s_position()[line.element_names.index('section_1_quad')] == 1
    assert line.elements[line.element_names.index('section_1_quad')] == elements['quad']
    assert line.get_s_position()[line.element_names.index('section_1_bend')] == 5
    assert line.elements[line.element_names.index('section_1_bend')] == elements['bend']

    assert line.get_s_position()[line.element_names.index('section_2')] == 10
    assert isinstance(line.elements[line.element_names.index('section_2')], xt.Marker)
    assert line.get_s_position()[line.element_names.index('section_2_quad')] == 11
    assert line.elements[line.element_names.index('section_2_quad')] == elements['quad']
    assert line.get_s_position()[line.element_names.index('section_2_bend')] == 15
    assert line.elements[line.element_names.index('section_2_bend')] == elements['bend']

    assert line.get_s_position()[line.element_names.index('sext')] == 13
    assert line.elements[line.element_names.index('sext')] == sext

    assert line.get_s_position()[line.element_names.index('section_3')] == 16
    assert isinstance(line.elements[line.element_names.index('section_3')], xt.Marker)
    assert line.get_s_position()[line.element_names.index('section_3_quad')] == 17
    assert line.elements[line.element_names.index('section_3_quad')] == elements['quad']

    # test negative drift
    # -------------------
    Line.from_sequence([Node(3, Multipole()), Node(2, Multipole())], 10, auto_reorder=True)

    with pytest.raises(ValueError):
        Line.from_sequence([Node(3, Multipole()), Node(2, Multipole())], 10)

    with pytest.raises(ValueError):
        Line.from_sequence([Node(1, Multipole()), Node(4, Multipole())], 2)


@pytest.mark.parametrize('refer', ['entry', 'centre', 'exit'])
def test_from_sequence_with_thick(refer):
    sequence = [
        xt.Node(1.2, xt.Drift(length=1), name='my_drift'),
        xt.Node(3, xt.Bend(length=1, k0=0.2), name='my_bend'),
    ]
    line = xt.Line.from_sequence(sequence, 5, refer=refer)  # noqa

    assert len(line) == 5
    assert line.get_length() == 5.0

    assert line.element_names[1] == 'my_drift'
    assert line.element_names[3] == 'my_bend'

    offset = 0
    if refer == 'centre':
        offset = -0.5
    elif refer == 'exit':
        offset = -1

    xo.assert_allclose(
        line.get_s_position(line.element_names),
        [
            0,             # drift
            1.2 + offset,  # my_drift
            2.2 + offset,  # drift
            3 + offset,    # my_bend
            4 + offset,    # drift
        ],
        atol=1e-15,
    )


def test_from_sequence_with_thick_fails():
    sequence = [
        xt.Node(1.2, xt.Drift(length=3), name='my_drift'),
        xt.Node(3, xt.Bend(length=3, k0=0.2), name='my_bend'),
    ]
    with pytest.raises(ValueError):
        _ = xt.Line.from_sequence(sequence, 5)


@for_all_test_contexts
def test_optimize_multipoles(test_context):
    elements = {
        'q1': xt.Multipole(knl=[0, 1], length=0, _context=test_context),
        'q2': xt.Multipole(knl=[1, 2], length=0, _context=test_context),
        'q3': xt.Multipole(knl=[0, 1], length=1, _context=test_context),
        'q4': xt.Multipole(knl=[0, 1], ksl=[1, 2], length=0, _context=test_context),
        'd1': xt.Multipole(knl=[1], hxl=0.0, length=2, _context=test_context),
        'd2': xt.Multipole(knl=[1], hxl=0.1, length=0, _context=test_context),
        'd4': xt.Multipole(knl=[1], ksl=[3], length=2, _context=test_context),
    }

    test_line = xt.Line(
        elements=elements,
        element_names=elements.keys(),
    )

    test_line.use_simple_bends()
    test_line.use_simple_quadrupoles()

    for nn in test_line.element_names:
        if nn in ('d1', 'd2'):
            assert type(test_line.element_dict[nn]) is xt.SimpleThinBend
        elif nn == 'q1' or nn == 'q3':
            assert type(test_line.element_dict[nn]) is xt.SimpleThinQuadrupole
        else:
            assert type(test_line.element_dict[nn]) is xt.Multipole

def test_from_json_to_json(tmp_path):

    line = xt.Line(
        elements={
            'm': xt.Multipole(knl=[1, 2]),
            'd': xt.Drift(length=1),
        },
        element_names=['m', 'd', 'm', 'd']
    )

    example_metadata = {
        'qx': {'lhcb1': 62.31, 'lhcb2': 62.31},
        'delta_cmr': 0.0,
    }
    line.metadata = example_metadata

    def asserts():
        assert len(result.element_dict.keys()) == 2
        assert result.element_names == ['m', 'd', 'm', 'd']

        assert isinstance(result['m'], xt.Multipole)
        assert (result['m'].knl == [1, 2]).all()

        assert isinstance(result['d'], xt.Drift)
        assert result['d'].length == 1

        assert result.metadata == example_metadata
        result.metadata['qx']['lhcb1'] = result.metadata['qx']['lhcb1'] + 1
        assert result.metadata != example_metadata
        result.metadata['qx']['lhcb1'] = result.metadata['qx']['lhcb1'] - 1

    line.to_json(tmp_path / 'test.json')
    result = xt.Line.from_json(tmp_path / 'test.json')

    asserts()

    with open(tmp_path / 'test2.json', 'w') as f:
        line.to_json(f)

    with open(tmp_path / 'test2.json', 'r') as f:
        result = xt.Line.from_json(f)

    asserts()

    with open(tmp_path / 'test2.json', 'w') as f:
        line.to_json(f,indent=None)

    with open(tmp_path / 'test2.json', 'r') as f:
        result = xt.Line.from_json(f)

    asserts()

    with open(tmp_path / 'test2.json.gz', 'w') as f:
        line.to_json(f,indent=2)

    with open(tmp_path / 'test2.json.gz', 'r') as f:
        result = xt.Line.from_json(f)

    asserts()


@for_all_test_contexts
def test_config_propagation(test_context):
    line = xt.Line(elements=10*[xt.Drift(length=1)])
    line.config.TEST1 = True
    line.config.TEST2 = 33.3
    line.matrix_stability_tol = 55.5
    dct = line.to_dict()

    assert 'config' in dct
    assert '_extra_config' in dct

    line2 = xt.Line.from_dict(dct)
    assert line2.config.TEST1 == True
    assert line2.config.TEST2 == 33.3
    assert line2.matrix_stability_tol == 55.5

    line2.build_tracker(_context=test_context)
    assert line2.tracker.matrix_stability_tol == 55.5

    # Check that they are copies
    line2.config.TEST1 = 23.3
    assert line2.config.TEST1 == 23.3
    assert line.config.TEST1 == True
    line2.matrix_stability_tol = 77.7
    assert line2.matrix_stability_tol == 77.7
    assert line.matrix_stability_tol == 55.5

    line3 = line.copy()
    assert line3.config.TEST1 == True
    assert line3.config.TEST2 == 33.3
    assert line3.matrix_stability_tol == 55.5

    # Check that they are copies
    line3.config.TEST1 = 23.3
    assert line3.config.TEST1 == 23.3
    assert line.config.TEST1 == True
    line3.matrix_stability_tol = 77.7
    assert line3.matrix_stability_tol == 77.7
    assert line.matrix_stability_tol == 55.5


def test_pickle():

    # Load the line
    line = xt.Line.from_json(test_data_folder /
            'hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
    line.particle_ref = xp.Particles(p0c=7e12, mass=xp.PROTON_MASS_EV)
    line.build_tracker()

    lnss = pickle.dumps(line)
    ln = pickle.loads(lnss)

    # Check that expressions work on old and new line
    line.vars['on_x1'] = 234
    ln.vars['on_x1'] = 123
    xo.assert_allclose(line.twiss(method='4d')['px', 'ip1'], 234e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(ln.twiss(method='4d')['px', 'ip1'], 123e-6, atol=1e-9, rtol=0)

    ln.vars['on_x1'] = 321
    xo.assert_allclose(line.twiss(method='4d')['px', 'ip1'], 234e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(ln.twiss(method='4d')['px', 'ip1'], 321e-6, atol=1e-9, rtol=0)

    line.vars['on_x1'] = 213
    xo.assert_allclose(line.twiss(method='4d')['px', 'ip1'], 213e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(ln.twiss(method='4d')['px', 'ip1'], 321e-6, atol=1e-9, rtol=0)

    line.discard_tracker()

    collider = xt.Environment(lines={'lhcb1': line})
    collider.build_trackers()

    colliderss = pickle.dumps(collider)

    coll = pickle.loads(colliderss)

    collider.vars['on_x1'] = 234
    coll.vars['on_x1'] = 123
    xo.assert_allclose(collider['lhcb1'].twiss(method='4d')['px', 'ip1'], 234e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(coll['lhcb1'].twiss(method='4d')['px', 'ip1'], 123e-6, atol=1e-9, rtol=0)

    coll.vars['on_x1'] = 321
    xo.assert_allclose(collider['lhcb1'].twiss(method='4d')['px', 'ip1'], 234e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(coll['lhcb1'].twiss(method='4d')['px', 'ip1'], 321e-6, atol=1e-9, rtol=0)

    collider.vars['on_x1'] = 213
    xo.assert_allclose(collider['lhcb1'].twiss(method='4d')['px', 'ip1'], 213e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(coll['lhcb1'].twiss(method='4d')['px', 'ip1'], 321e-6, atol=1e-9, rtol=0)


def test_line_attr():
    line = xt.Line(
        elements=[
            xt.Drift(length=1),
            xt.Multipole(knl=[2, 3, 4], hxl=8),
            xt.Bend(k0=5, h=0.5, length=6, knl=[7, 8, 9]),
            xt.Drift(length=10),
            xt.Quadrupole(k1=11, length=12),
        ]
    )

    line.build_tracker()

    assert np.all(line.attr['length'] == [1, 0, 6, 10, 12])
    assert np.all(line.attr['k0l'] == [0, 2, 5 * 6 + 7, 0, 0])
    assert np.all(line.attr['k1l'] == [0, 3, 8, 0, 11 * 12])
    assert np.all(line.attr['angle_rad'] == [0, 8, 0.5 * 6, 0, 0])

@for_all_test_contexts
def test_insert_thin_elements_at_s_basic(test_context):

    l1 = xt.Line(elements=5*[xt.Drift(length=1)])
    l1.build_tracker(_context=test_context) # Move all elements to selected context

    l1.discard_tracker()
    l1._insert_thin_elements_at_s([
        (0, [(f'm0_at_a', xt.Marker(_context=test_context)), (f'm1_at_a', xt.Marker(_context=test_context))]),
        (5, [(f'm0_at_b', xt.Marker(_context=test_context)), (f'm1_at_b', xt.Marker(_context=test_context))]),
    ])

    t1 = l1.get_table()
    assert t1.name[0] == 'm0_at_a'
    assert t1.name[1] == 'm1_at_a'
    assert t1.name[-1] == '_end_point'
    assert t1.name[-2] == 'm1_at_b'
    assert t1.name[-3] == 'm0_at_b'

    assert t1.s[0] == 0
    assert t1.s[1] == 0
    assert t1.s[-1] == 5.
    assert t1.s[-2] == 5.
    assert t1.s[-3] == 5.

@for_all_test_contexts
def test_insert_thin_elements_at_s_lhc(test_context):

    line = xt.Line.from_json(test_data_folder /
                    'hllhc15_thick/lhc_thick_with_knobs.json')
    line.twiss_default['method'] = '4d'

    line.build_tracker(_context=test_context)

    Strategy = xt.Strategy
    Teapot = xt.Teapot
    slicing_strategies = [
        Strategy(slicing=Teapot(1)),  # Default catch-all as in MAD-X
        Strategy(slicing=Teapot(4), element_type=xt.Bend),
        Strategy(slicing=Teapot(20), element_type=xt.Quadrupole),
        Strategy(slicing=Teapot(2), name=r'^mb\..*'),
        Strategy(slicing=Teapot(5), name=r'^mq\..*'),
        Strategy(slicing=Teapot(2), name=r'^mqt.*'),
        Strategy(slicing=Teapot(60), name=r'^mqx.*'),
    ]

    line.discard_tracker()
    line.slice_thick_elements(slicing_strategies=slicing_strategies)

    tw0 = line.twiss()
    line.discard_tracker()

    e0 = 'mq.28r3.b1_entry'
    e1 = 'mq.29r3.b1_exit'

    s0 = line.get_s_position(e0)
    s1 = line.get_s_position(e1)
    s2 = line.get_length()

    elements_to_insert = [
        # s .    # elements to insert (name, element)
        (s0,     [(f'm0_at_a', xt.Marker(_context=test_context)), (f'm1_at_a', xt.Marker(_context=test_context)), (f'm2_at_a', xt.Marker(_context=test_context))]),
        (s0+10., [(f'm0_at_b', xt.Marker(_context=test_context)), (f'm1_at_b', xt.Marker(_context=test_context)), (f'm2_at_b', xt.Marker(_context=test_context))]),
        (s1,     [(f'm0_at_c', xt.Marker(_context=test_context)), (f'm1_at_c', xt.Marker(_context=test_context)), (f'm2_at_c', xt.Marker(_context=test_context))]),
        (s2,     [(f'm0_at_d', xt.Marker(_context=test_context)), (f'm1_at_d', xt.Marker(_context=test_context)), (f'm2_at_d', xt.Marker(_context=test_context))]),
    ]

    line.discard_tracker()
    line._insert_thin_elements_at_s(elements_to_insert)
    line.build_tracker(_context=test_context)

    tt = line.get_table()

    # Check that there are no duplicated elements
    assert len(tt.name) == len(set(tt.name))

    xo.assert_allclose(tt['s', 'm0_at_a'], s0, rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm1_at_a'], s0, rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm2_at_a'], s0, rtol=0, atol=1e-6)

    xo.assert_allclose(tt['s', 'm0_at_b'], s0 + 10., rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm1_at_b'], s0 + 10., rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm2_at_b'], s0 + 10., rtol=0, atol=1e-6)

    xo.assert_allclose(tt['s', 'm0_at_c'], s1, rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm1_at_c'], s1, rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm2_at_c'], s1, rtol=0, atol=1e-6)

    xo.assert_allclose(tt['s', 'm0_at_d'], s2, rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm1_at_d'], s2, rtol=0, atol=1e-6)
    xo.assert_allclose(tt['s', 'm2_at_d'], s2, rtol=0, atol=1e-6)

    assert np.all(tt.rows['mq.28r3.b1_entry<<3':'mq.28r3.b1_entry'].name
            == np.array(['m0_at_a', 'm1_at_a', 'm2_at_a', 'mq.28r3.b1_entry']))

    assert np.all(tt.rows['m0_at_b<<2':'m0_at_b>>4'].name
            == np.array(['mb.a29r3.b1..0', 'drift_mb.a29r3.b1..1..0',
                        'm0_at_b', 'm1_at_b', 'm2_at_b',
                        'drift_mb.a29r3.b1..1..1', 'mb.a29r3.b1..1']))

    assert np.all(tt.rows['mq.29r3.b1_exit<<3':'mq.29r3.b1_exit'].name
            == np.array(
                ['m1_at_c', 'm2_at_c', 'mq.29r3.b1..exit_map', 'mq.29r3.b1_exit']))

    assert np.all(tt.rows['m0_at_d':'m0_at_d>>4'].name
                == np.array(['m0_at_d', 'm1_at_d', 'm2_at_d',
                            'lhcb1ip7_p_', '_end_point']))

    xo.assert_allclose(line.get_length(), tw0.s[-1], atol=1e-6)

    tw1 = line.twiss()
    xo.assert_allclose(tw1.qx, tw0.qx, atol=1e-9, rtol=0)


def test_elements_intersecting_s():
    elements = {
        'e1': xt.Drift(length=1),  # at 0
        'e2': xt.Drift(length=2),  # at 1
        'm1': xt.Marker(),         # at 3
        'e3': xt.Drift(length=1),  # at 3
        'm2': xt.Marker(),         # at 4
        'm3': xt.Marker(),         # at 4
        'e4': xt.Drift(length=1),  # at 4
        'e5': xt.Drift(length=2),  # at 5
        'e6': xt.Drift(length=1),  # at 7
        'm4': xt.Marker(),         # at 8
    }

    line = xt.Line(
        elements=elements,
        element_names=list(elements.keys()),
    )

    cuts = [
        0.5,  # e1
        1,
        1.1, 1.7,  # e2
        4.5,  # e4
        7.8, 7.9,  # e6
        8,
        8,
    ]

    expected = {
        'e1': [0.5],
        'e2': [0.1, 0.7],
        'e4': [0.5],
        'e6': [0.8, 0.9],
    }
    result = line._elements_intersecting_s(cuts)
    for kk in expected.keys() | result.keys():
        xo.assert_allclose(expected[kk], result[kk], atol=1e-16)


def test_slicing_at_custom_s():
    elements = {
        'e1': xt.Drift(length=1),  # at 0
        'e2': xt.Drift(length=2),  # at 1
        'm1': xt.Marker(),         # at 3
        'e3': xt.Drift(length=1),  # at 3
        'm2': xt.Marker(),         # at 4
        'm3': xt.Marker(),         # at 4
        'e4': xt.Drift(length=1),  # at 4
        'e5': xt.Drift(length=2),  # at 5
        'e6': xt.Drift(length=1),  # at 7
        'm4': xt.Marker(),         # at 8
    }

    line = xt.Line(
        elements=elements,
        element_names=list(elements.keys()),
    )

    cuts = [
        0.5,  # e1
        1,
        1.1, 1.7,  # e2
        4.5,  # e4
        7.8, 7.9,  # e6
        8,
        8,
    ]

    line.cut_at_s(cuts)

    tab = line.get_table()
    xo.assert_allclose(tab.rows[r'e1\.\.\d*'].s, [0, 0.5], atol=1e-16)
    xo.assert_allclose(tab.rows[r'e2\.\.\d*'].s, [1, 1.1, 1.7], atol=1e-16)
    xo.assert_allclose(tab.rows[r'e3'].s, [3], atol=1e-16)
    xo.assert_allclose(tab.rows[r'e4\.\.\d*'].s, [4, 4.5], atol=1e-16)
    xo.assert_allclose(tab.rows[r'e5'].s, [5], atol=1e-16)
    xo.assert_allclose(tab.rows[r'e6\.\.\d*'].s, [7, 7.8, 7.9], atol=1e-16)

def test_insert_thick_element_reuse_marker_name():

    assert_allclose = xo.assert_allclose

    elements = {
        'd1': xt.Drift(length=1),
        'm1': xt.Marker(),
        'd2': xt.Drift(length=1),
    }

    line=xt.Line(elements=elements,
                element_names=list(elements.keys()))

    # Note that the name is reused
    line.insert_element(element=xt.Bend(length=1.), name='m1', at_s=0.5)

    tt = line.get_table()

    assert np.all(tt.name == ['d1..0', 'm1', 'd2..1', '_end_point'])
    assert np.all(tt.parent_name == ['d1', None, 'd2', None])
    assert_allclose(tt.s, [0. , 0.5, 1.5, 2. ], rtol=0, atol=1e-14)

def test_multiple_thick_elements():
    line = xt.Line(
        elements=[xt.Drift(length=1.0) for i in range(10)]
    )

    s_insert = np.array([2.5, 5.5, 7])
    l_insert = np.array([1.0, 1.0, 1.0])
    ele_insert = [xt.Sextupole(length=l) for l in l_insert]

    line._insert_thick_elements_at_s(
        element_names=[f'insertion_{i}' for i in range(len(s_insert))],
        elements=ele_insert,
        at_s=s_insert
    )

    tt = line.get_table()

    assert np.all(tt.name == ['e0', 'e1', 'e2..0', 'insertion_0', 'e3..1', 'e4', 'e5..0',
        'insertion_1', 'e6..1', 'insertion_2', 'e8', 'e9', '_end_point'])
    xo.assert_allclose(tt.s, [ 0. ,  1. ,  2. ,  2.5,  3.5,  4. ,  5. ,
                                    5.5,  6.5,  7. ,  8. , 9. , 10. ])

    assert np.all(tt.element_type == ['Drift', 'Drift', 'DriftSlice', 'Sextupole', 'DriftSlice', 'Drift',
        'DriftSlice', 'Sextupole', 'DriftSlice', 'Sextupole', 'Drift',
        'Drift', ''])

@for_all_test_contexts
def test_get_strengths(test_context):
    collider = xt.Environment.from_json(
        test_data_folder / 'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers(_context=test_context)

    collider.lhcb1.twiss_default['method'] = '4d'
    collider.lhcb2.twiss_default['method'] = '4d'
    collider.lhcb2.twiss_default['reverse'] = True

    line = collider.lhcb2 # <- use lhcb2 to test the reverse option

    import xobjects as xo
    str_table_rev = line.get_strengths() # Takes reverse from twiss_default
    xo.assert_allclose(line['mbw.a6l3.b2'].k0,
            -str_table_rev['k0l', 'mbw.a6l3.b2'] / str_table_rev['length', 'mbw.a6l3.b2'],
            rtol=0, atol=1e-14)
    xo.assert_allclose(line['mbw.a6l3.b2'].h,
            -str_table_rev['angle_rad', 'mbw.a6l3.b2'] / str_table_rev['length', 'mbw.a6l3.b2'],
            rtol=0, atol=1e-14)

    str_table = line.get_strengths(reverse=False) # Takes reverse from twiss_default
    xo.assert_allclose(line['mbw.a6l3.b2'].k0,
            str_table['k0l', 'mbw.a6l3.b2'] / str_table['length', 'mbw.a6l3.b2'],
            rtol=0, atol=1e-14)
    xo.assert_allclose(line['mbw.a6l3.b2'].h,
            str_table['angle_rad', 'mbw.a6l3.b2'] / str_table['length', 'mbw.a6l3.b2'],
            rtol=0, atol=1e-14)



def test_insert_repeated_names():

    line = xt.Line(
        elements=([xt.Drift(length=0)] # Start line marker
                    + [xt.Drift(length=1) for _ in range(5)]
                    + [xt.Drift(length=0)] # End line marker
            ),
        element_names=['d']*7
        )
    line.insert_element("m1",xt.Marker(),at="d::3")
    assert line.element_names[3]=="m1"
    line.insert_element("m2",xt.Marker(),at="d")
    assert line.element_names[0]=="m2"

def test_line_table_unique_names():
    line = xt.Line(
        elements = {"obm": xt.Bend(length=0.5)},
        element_names= ["obm","obm"]
    )
    table = line.get_table()
    names, counts = np.unique(table.name, return_counts=True, equal_nan=False)
    assert np.all(counts == 1), "Not all elements are unique"
    for name, env_name in zip(table.name, table.env_name):
        if name == '_end_point': continue
        assert line[name] == line[env_name]


def test_extend_knl_ksl():

    classes_to_check = ['Bend', 'Quadrupole', 'Sextupole', 'Octupole', 'Solenoid',
                        'Multipole']

    for cc in classes_to_check:

        nn1 = 'test1_'+cc.lower()
        nn2 = 'test2_'+cc.lower()
        env = xt.Environment()
        env.new(nn1, cc, length=10, knl=[
                1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], ksl=[3, 2, 1])
        env.new(nn2, cc, length=10, ksl=[
                1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], knl=[3, 2, 1], order=11)

        assert env[nn1].__class__.__name__ == cc
        assert env[nn1].order == 11
        assert len(env[nn1].knl) == 12
        assert len(env[nn1].ksl) == 12
        xo.assert_allclose(env[nn1].knl, [1, 2, 3, 4, 5,
                           6, 7, 8, 9, 10, 11, 12], rtol=0, atol=1e-15)
        xo.assert_allclose(env[nn1].ksl, [3, 2, 1, 0, 0,
                           0, 0, 0, 0, 0, 0, 0], rtol=0, atol=1e-15)
        xo.assert_allclose(env[nn1].inv_factorial_order,
                           1/math.factorial(11), rtol=0, atol=1e-15)

        assert env[nn2].__class__.__name__ == cc
        assert env[nn2].order == 11
        assert len(env[nn2].ksl) == 12
        assert len(env[nn2].knl) == 12
        xo.assert_allclose(env[nn2].ksl, [1, 2, 3, 4, 5,
                           6, 7, 8, 9, 10, 11, 12], rtol=0, atol=1e-15)
        xo.assert_allclose(env[nn2].knl, [3, 2, 1, 0, 0,
                           0, 0, 0, 0, 0, 0, 0], rtol=0, atol=1e-15)
        xo.assert_allclose(env[nn2].inv_factorial_order,
                           1/math.factorial(11), rtol=0, atol=1e-15)

    env.vars.default_to_zero = True
    line = env.new_line(components=[
        env.new('b1', xt.Bend, length=1, knl=[
                'a', 'b', 'c'], ksl=['d', 'e', 'f']),
        env.new('q1', xt.Quadrupole, length=1, knl=[
                'a', 'b', 'c'], ksl=['d', 'e', 'f']),
        env.new('s1', xt.Sextupole, length=1, knl=[
                'a', 'b', 'c'], ksl=['d', 'e', 'f']),
        env.new('o1', xt.Octupole, length=1, knl=[
                'a', 'b', 'c'], ksl=['d', 'e', 'f']),
        env.new('s2', xt.Solenoid, length=1, knl=[
                'a', 'b', 'c'], ksl=['d', 'e', 'f']),
        env.new('m1', xt.Multipole, length=1, knl=[
                'a', 'b', 'c'], ksl=['d', 'e', 'f']),
    ])

    env['a'] = 3.
    env['b'] = 2.
    env['c'] = 1.
    env['d'] = 4.
    env['e'] = 5.
    env['f'] = 6.

    element_names = ['b1', 'q1']
    order = 10

    line.extend_knl_ksl(order=order, element_names=element_names)

    assert line['b1'].order == order
    assert line['q1'].order == order
    assert line['s1'].order == 5
    assert line['o1'].order == 5
    assert line['s2'].order == 5
    assert line['m1'].order == 2

    xo.assert_allclose(line['b1'].inv_factorial_order,
                       1/math.factorial(order), rtol=0, atol=1e-15)
    xo.assert_allclose(line['q1'].inv_factorial_order,
                       1/math.factorial(order), rtol=0, atol=1e-15)
    xo.assert_allclose(line['s1'].inv_factorial_order,
                       1/math.factorial(5), rtol=0, atol=1e-15)
    xo.assert_allclose(line['o1'].inv_factorial_order,
                       1/math.factorial(5), rtol=0, atol=1e-15)
    xo.assert_allclose(line['s2'].inv_factorial_order,
                       1/math.factorial(5), rtol=0, atol=1e-15)
    xo.assert_allclose(line['m1'].inv_factorial_order,
                       1/math.factorial(2), rtol=0, atol=1e-15)

    xo.assert_allclose(line['b1'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['b1'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['q1'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['q1'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s1'].knl, [3., 2., 1.,
                       0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s1'].ksl, [4., 5., 6.,
                       0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['o1'].knl, [3., 2., 1.,
                       0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['o1'].ksl, [4., 5., 6.,
                       0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s2'].knl, [3., 2., 1.,
                       0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s2'].ksl, [4., 5., 6.,
                       0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['m1'].knl, [3., 2., 1.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['m1'].ksl, [4., 5., 6.], rtol=0, atol=1e-15)

    line.extend_knl_ksl(order=11)

    assert line['b1'].order == 11
    assert line['q1'].order == 11
    assert line['s1'].order == 11
    assert line['o1'].order == 11
    assert line['s2'].order == 11
    assert line['m1'].order == 11
    assert line['b1'].inv_factorial_order == 1/math.factorial(11)
    assert line['q1'].inv_factorial_order == 1/math.factorial(11)
    assert line['s1'].inv_factorial_order == 1/math.factorial(11)
    assert line['o1'].inv_factorial_order == 1/math.factorial(11)
    assert line['s2'].inv_factorial_order == 1/math.factorial(11)
    assert line['m1'].inv_factorial_order == 1/math.factorial(11)
    xo.assert_allclose(line['b1'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['b1'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['q1'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['q1'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s1'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s1'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['o1'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['o1'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s2'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['s2'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['m1'].knl, [3., 2., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['m1'].ksl, [4., 5., 6., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)

    # test an expression
    line['b'] = 100
    line['f'] = 200

    xo.assert_allclose(line['o1'].knl, [3., 100., 1., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)
    xo.assert_allclose(line['o1'].ksl, [4., 5., 200., 0.,
                       0., 0., 0., 0., 0., 0., 0., 0.], rtol=0, atol=1e-15)





import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment(
    particle_ref=xt.Particles(p0c=7000e9, x=1e-3, px=1e-3, y=1e-3, py=1e-3))

line0 = env.new_line(
    components=[
        env.new('b0', 'Bend', length=1.0, anchor='start', at=5.0),
        env.new('q0', 'Quadrupole', length=2.0, at=20.0),
        env.new('ql', 'Quadrupole', length=2.0, at=-10.0, from_='q0'),
        env.new('qr', 'Quadrupole', length=2.0, at=10.0, from_='q0'),
        env.new('mk1', 'Marker', at=40),
        env.new('mk2', 'Marker'),
        env.new('mk3', 'Marker'),
        env.place('q0'),
        env.place('b0'),
        env.new('end', 'Marker', at=50.),
    ])

line = line0.copy()
line.env.new('qnew', 'Quadrupole', length=2.0)
line.append('qnew')

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'b0::0', 'drift_2', 'ql', 'drift_3', 'q0::0', 'drift_4',
       'qr', 'drift_5', 'mk1', 'mk2', 'mk3', 'q0::1', 'b0::1', 'drift_6',
       'end', 'qnew', '_end_point']))
xo.assert_allclose(tt.s_center, np.array(
    [ 2.5,  5.5,  7.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. ,
       40. , 41. , 42.5, 46.5, 50. , 51. , 52. ]),
    rtol=0., atol=1e-14)

class MyElement:
    def __init__(self, myparameter):
        self.myparameter = myparameter

    def track(self, particles):
        particles.px += self.myparameter

myelem = MyElement(0.1)

line = line0.copy()
line.append('myname', myelem)

tt = line.get_table()
tt.show(cols=['name', 'element_type', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'b0::0', 'drift_2', 'ql', 'drift_3', 'q0::0', 'drift_4',
       'qr', 'drift_5', 'mk1', 'mk2', 'mk3', 'q0::1', 'b0::1', 'drift_6',
       'end', 'myname', '_end_point']))

assert np.all(tt.element_type == np.array(
    ['Drift', 'Bend', 'Drift', 'Quadrupole', 'Drift', 'Quadrupole',
       'Drift', 'Quadrupole', 'Drift', 'Marker', 'Marker', 'Marker',
       'Quadrupole', 'Bend', 'Drift', 'Marker', 'MyElement', '']))

xo.assert_allclose(tt.s_center, np.array(
    [ 2.5,  5.5,  7.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. ,
       40. , 41. , 42.5, 46.5, 50. , 50. , 50. ]),
    rtol=0., atol=1e-14)

line = line0.copy()
line.env.new('qnew1', 'Quadrupole', length=2.0)
line.env.new('qnew2', 'Quadrupole', length=2.0)
line.append(['qnew1', 'qnew2'])

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'b0::0', 'drift_2', 'ql', 'drift_3', 'q0::0', 'drift_4',
       'qr', 'drift_5', 'mk1', 'mk2', 'mk3', 'q0::1', 'b0::1', 'drift_6',
       'end', 'qnew1', 'qnew2', '_end_point']))
xo.assert_allclose(tt.s_center, np.array(
    [ 2.5,  5.5,  7.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. ,
       40. , 41. , 42.5, 46.5, 50. , 51. , 53. , 54. ]),
    rtol=0., atol=1e-14)

line = line0.copy()
line.env.new('qnew1', 'Quadrupole', length=2.0)
line.env.new('qnew2', 'Quadrupole', length=2.0)

l2 = line.env.new_line(components=['qnew1', 'qnew2', 'ql'])
line.append(l2)

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'b0::0', 'drift_2', 'ql::0', 'drift_3', 'q0::0', 'drift_4',
       'qr', 'drift_5', 'mk1', 'mk2', 'mk3', 'q0::1', 'b0::1', 'drift_6',
       'end', 'qnew1', 'qnew2', 'ql::1', '_end_point']))

xo.assert_allclose(tt.s_center, np.array(
    [ 2.5,  5.5,  7.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. ,
       40. , 41. , 42.5, 46.5, 50. , 51. , 53. , 55. , 56. ]),
    rtol=0., atol=1e-14)

line = line0.copy()
line.env.new('qnew1', 'Quadrupole', length=2.0)
line.env.new('qnew2', 'Quadrupole', length=2.0)

l2 = line.env.new_line(components=['qnew1', 'qnew2', 'ql'])
line.append([l2, 'qr'])

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'b0::0', 'drift_2', 'ql::0', 'drift_3', 'q0::0', 'drift_4',
       'qr::0', 'drift_5', 'mk1', 'mk2', 'mk3', 'q0::1', 'b0::1', 'drift_6',
       'end', 'qnew1', 'qnew2', 'ql::1', 'qr::1', '_end_point']))
xo.assert_allclose(tt.s_center, np.array(
    [ 2.5,  5.5,  7.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. ,
       40. , 41. , 42.5, 46.5, 50. , 51. , 53. , 55. , 57. , 58. ]),
    rtol=0., atol=1e-14)



from typing import Dict, Optional, List, Set, Tuple, Union

import numpy as np

import xtrack as xt
from xtrack import BeamElement
from xtrack.environment import Builder
from xtrack.mad_parser.parse import ElementType, LineType, MadxParser, VarType, MadxOutputType
from xtrack.environment import _reverse_element

EXTRA_PARAMS = {
    "slot_id",
    "mech_sep",
    "assembly_id",
    "kmin",
    "kmax",
    "calib",
    "polarity",
    "aper_tol",
}

TRANSLATE_PARAMS = {
    "l": "length",
    "lrad": "length",
    "tilt": "rot_s_rad",
    "from": "from_",
    "e1": "edge_entry_angle",
    "e2": "edge_exit_angle",
    "fint": "edge_entry_fint",
    "fintx": "edge_exit_fint",
}

CONSTANTS = {
    "pi": np.pi,
    "twopi": np.pi * 2,
    "degrad": 180 / np.pi,  # deg/rad
    "raddeg": np.pi / 180,  # rad/deg
    "e": np.e,
    "emass": 0.51099895000e-3,  # GeV
    "pmass": 0.93827208816,  # GeV
    "nmass": 0.93956542052,  # GeV
    "umass": 0.93149410242,  # GeV
    "mumass": 0.1056583715,  # GeV
    "clight": 299792458.0,  # m/s
    "qelect": 1.602176634e-19,  # A * s
    "hbar": 6.582119569e-25,  # MeV * s
    "erad": 2.8179403262e-15,  # m
    "prad": 'erad / emass * pmass',
}

_APERTURE_TYPES = {
    'circle': 'LimitEllipse',
    'ellipse': 'LimitEllipse',
    'rectangle': 'LimitRect',
    'rectellipse': 'LimitRectEllipse',
    'racetrack': 'LimitRacetrack',
    'octagon': 'LimitPolygon',
    'polygon': 'LimitPolygon',  # not really an explicit option in MAD-X
}


def _warn(msg):
    print(f'Warning: {msg}')


def get_params(params, parent):
    params = params.copy()
    if parent in {'placeholder', 'instrument'}:
        _ = params.pop('lrad', None)

    def _normalise_single(param):
        lower = param.lower()
        return TRANSLATE_PARAMS.get(lower, lower)

    # TODO: Ignoring value['deferred'] for now.
    normalised = {_normalise_single(k): v for k, v in params.items()}

    main_params = {}
    extras = {}
    for k, v in normalised.items():
        if k in EXTRA_PARAMS:
            extras[k] = v
        else:
            main_params[k] = v

    return main_params, extras


class MadxLoader:
    def __init__(
            self,
            env: xt.Environment = None,
            default_to_zero: bool = False,
    ):
        self._madx_elem_hierarchy: Dict[str, List[str]] = {}
        self._both_direction_elements: Set[str] = set()
        self._builtin_types = set()
        self._parameter_cache = {}

        self.rbarc = True

        self.env = env or xt.Environment()
        self.env.default_to_zero = default_to_zero

        self._init_environment()

    def _init_environment(self):
        self.env.vars.default_to_zero = True

        # Define the builtin MAD-X variables
        self.env.vars.update(CONSTANTS)

        # Define the built-in MAD-X elements
        self._new_builtin("vkicker", "Multipole")
        self._new_builtin("hkicker", "Multipole")
        self._new_builtin("tkicker", "Multipole")
        self._new_builtin("kicker", "Multipole")
        self._new_builtin("drift", "Drift")
        self._new_builtin("collimator", "Drift")
        self._new_builtin("rcollimator", "Drift")
        self._new_builtin("ecollimator", "Drift")
        self._new_builtin("instrument", "Drift")
        self._new_builtin("monitor", "Drift")
        self._new_builtin("hmonitor", "Drift")
        self._new_builtin("vmonitor", "Drift")
        self._new_builtin("placeholder", "Drift")
        self._new_builtin("sbend", "Bend")
        self._new_builtin("rbend", "RBend")
        self._new_builtin("quadrupole", "Quadrupole")
        self._new_builtin("sextupole", "Sextupole")
        self._new_builtin("octupole", "Octupole")
        self._new_builtin("marker", "Marker")
        self._new_builtin("rfcavity", "Cavity")
        self._new_builtin("multipole", "Multipole", knl=6 * [0])
        self._new_builtin("solenoid", "Solenoid")

        for mad_apertype in _APERTURE_TYPES:
            self._new_builtin(mad_apertype, 'Marker')

    def load_file(self, file, build=True) -> Optional[List[Builder]]:
        """Load a MAD-X file and generate/update the environment."""
        parser = MadxParser(vars=self.env.vars, functions=self.env.functions)
        parsed_dict = parser.parse_file(file)
        return self.load_parsed_dict(parsed_dict, build=build)

    def load_string(self, string, build=True) -> Optional[List[Builder]]:
        """Load a MAD-X string and generate/update the environment."""
        parser = MadxParser(vars=self.env.vars, functions=self.env.functions)
        parsed_dict = parser.parse_string(string)
        return self.load_parsed_dict(parsed_dict, build=build)

    def load_parsed_dict(self, parsed_dict: MadxOutputType, build=True) -> Optional[List[Builder]]:
        hierarchy = self._collect_hierarchy(parsed_dict)
        self._madx_elem_hierarchy.update(hierarchy)

        self._parse_elements(parsed_dict["elements"])
        builders = self._parse_lines(parsed_dict["lines"], build=build)
        self._parse_parameters(parsed_dict["parameters"])

        if not build:
            return builders

    def _parse_elements(self, elements: Dict[str, ElementType]):
        for name, el_params in elements.items():
            parent = el_params.pop('parent')
            assert parent != 'sequence'
            params, extras = get_params(el_params, parent=parent)
            self._new_element(name, parent, self.env, **params, extra=extras)

    def _parse_lines(self, lines: Dict[str, LineType], build=True) -> List[Builder]:
        builders = []

        for name, line_params in lines.items():
            params = line_params.copy()
            line_type = params.pop('parent')
            self._madx_elem_hierarchy[name] = [line_type]

            if line_type == 'sequence':
                refer = params.get('refer', 'centre')
                if refer == 'entry':
                    refer = 'start'
                elif refer == 'exit':
                    refer = 'end'
                length = params.get('l', None)
                builder = self.env.new_builder(name=name, refer=refer,
                                               length=length)
                self._parse_components(builder, params.pop('elements'))
                builders.append(builder)
            elif line_type == 'line':
                components = self._parse_line_components(params.pop('elements'))
                builder = self.env.new_builder(name=name, components=components)
            else:
                raise ValueError(
                    f'Only a MAD-X sequence or a line type can be used to build'
                    f'a line, but got: {line_type}!'
                )

            if build:
                builder.build()

        return builders

    def _parse_parameters(self, parameters: Dict[str, Dict[str, str]]):
        for element, el_params in parameters.items():
            params, extras = get_params(el_params, parent=element)
            self._set_element(element, self.env, **params, extra=extras)

    def _parse_components(self, builder, elements: List[Tuple[str, Union[ElementType, LineType]]]):
        for name, element in elements:
            params = element.copy()
            parent = params.pop('parent', None)
            assert parent != 'sequence'
            params, extras = get_params(params, parent=parent)
            self._new_element(name, parent, builder, **params, extra=extras)

    def _parse_line_components(self, elements):
        components = []

        for name, body in elements:
            # Parent is None if the element already exists and is referred to,
            # by name, otherwise we expect a line nested in the current one.
            parent = body.get('parent', None)
            repeat = body.get('_repeat', 1)
            invert = body.get('_invert', False)
            instance = self.env[name] if name else None

            if parent is None and isinstance(instance, xt.Line):
                # If it's a line, we use __mul__ and __neg__ directly
                element = instance
                if invert:
                    element = -element
                element = repeat * element
                components.append(element)
            elif parent == 'line':
                # If it's a nested line, we parse it recursively
                element = self._parse_line_components(body['elements'])
                if invert:
                    element = list(reversed(element))
                element = repeat * element
                components += element
            elif parent is None:
                # If it's a reference to a single element, we multiply it and
                # add it. Reversal will not affect it.
                components += body.get('_repeat', 1) * [name]
            else:
                raise ValueError('Only an element reference or a line is accepted')

        return components

    def _new_builtin(self, name, xt_type, **kwargs):
        if name not in self.env.element_dict:
            self.env.new(name, xt_type, **kwargs)
        self._builtin_types.add(name)

    def _new_element(self, name, parent, builder, **kwargs):
        should_clone = parent is not None

        if should_clone:
            self._parameter_cache[name] = self._parameter_cache.get(parent, {}).copy()
        else:
            self._parameter_cache[name] = self._parameter_cache.get(name, {})
        self._parameter_cache[name].update(kwargs)

        aperture = self._build_aperture(name, f'{name}_aper', kwargs)
        if not aperture and name in self.env.element_dict:
            aperture = self.env[name].name_associated_aperture
        if not aperture and parent:
            aperture = self.env[parent].name_associated_aperture

        el_params = self._convert_element_params(name, kwargs)

        if aperture and 'at' in el_params:  # placing mode
            builder.place(aperture, at=0, from_=f'{name}@start')

        if should_clone:
            self._clone_element(name, parent, builder, el_params)
        else:
            # If parent is None, we must be in a sequence, and so we are
            # placing the element: in MAD-X this requires an `at` param, but
            # we can be a bit more lax, as Xsuite will automatically place the
            # element after the previous one if there is not `at`.
            self._place_element(name, el_params, builder)

        if aperture:
            builder.element_dict[name].name_associated_aperture = aperture

    def _place_element(self, name, el_params, builder):
        """Place an element in the sequence.

        This is the case when `parent` is None.
        """
        if self._mad_base_type(name) in ['rbend', 'sbend']:
            el_params.pop('k0_from_h', None)

        if (superfluous := el_params.keys() - {'at', 'from_', 'extra'}):
            raise ValueError(
                f'Cannot place the element `{name}` as it overrides the '
                f'parameters: {superfluous}!'
            )

        if (extras := el_params.pop('extra', None)):
            _warn(f'Ignoring extra parameters {extras} for element `{name}`!')
        element = self.env[name]

        length = self._element_length(name, el_params)
        is_not_thick = isinstance(element, BeamElement) and not element.isthick
        if is_not_thick and length and not isinstance(element, xt.Marker):
            # Handle the thin elements that have a length in MAD-X: sandwich
            line = self._make_thick_sandwich(name, length)
            builder.place(line, **el_params)
        else:
            builder.place(name, **el_params)

    def _clone_element(self, name, parent, builder, el_params):
        """Clone an element, and possibly place it if we are in a sequence.

        Here `parent` is not None.
        """
        length = self._element_length(name, el_params)
        element = self.env[parent]
        is_not_thick = isinstance(element, BeamElement) and not element.isthick

        if is_not_thick and length and not isinstance(element, xt.Marker):
            # Handle the thin elements that have a length in MAD-X: sandwich
            at, from_ = el_params.pop('at', None), el_params.pop('from_', None)
            if name not in self.env.element_dict:
                make_drifts = True
                self.env.new(name, parent, **el_params)
            else:
                make_drifts = False
                _warn(f'Element `{name}` already exists, this definition '
                      f'will be ignored (for compatibility with MAD-X)')
            name = self._make_thick_sandwich(name, length, make_drifts)
            builder.place(name, at=at, from_=from_)
        elif name == parent:
            # This happens when the element is cloned with the same name, which
            # is allowed inside MAD-X sequences, e.g.: `el: el, at = 42;`.
            # We cannot attach extra to a place though, so this is skipped.
            if dropped_extra := el_params.pop('extra', None):
                _warn(f'Ignoring extra parameters {dropped_extra} for element '
                      f'`{name}`: it is a clone of itself overriding `extra`.')
            el_params.pop('k0_from_h', None)
            builder.place(name, **el_params)
        else:
            # `force=True` is needed to overwrite existing elements. In MAD-X
            # when an element name is repeated between lines, the first one
            # is retained: we do not simulate this behaviour here.
            builder.new(name, parent, force=True, **el_params)

    def _element_length(self, name, el_params):
        """Given the definition and params of the element, return its length."""
        length = el_params.get('length', self._parameter_cache[name].get('length', 0))
        THIN_ELEMENTS = {'vkicker', 'hkicker', 'kicker', 'tkicker', 'multipole'}
        if self._mad_base_type(name) in THIN_ELEMENTS:
            # Workaround for the elements that are thin despite having a
            # ``length`` parameter.
            length = 0

        return length

    def _make_thick_sandwich(self, name, length, make_drifts=True):
        """Make a sandwich of two drifts around the element."""
        drift_name = f'{name}_drift'
        if make_drifts:
            self.env.new(drift_name + '_0', 'Drift', length=length / 2)
            self.env.new(drift_name + '_1', 'Drift', length=length / 2)
        line = self.env.new_line([drift_name + '_0', name, drift_name + '_1'])
        return line

    def _set_element(self, name, builder, **kwargs):
        self._parameter_cache[name].update(kwargs)

        if 'aperture' in kwargs and 'apertype' not in kwargs:
            kwargs['apertype'] = self._parameter_cache[name]['apertype']
        aperture = self._build_aperture(name, f'{name}_aper', kwargs, force=True)

        el_params = self._convert_element_params(name, kwargs)
        builder.set(name, **el_params)
        builder.element_dict[name].name_associated_aperture = aperture

    def _convert_element_params(self, name, params):
        parent_name = self._mad_base_type(name)

        if parent_name in {'sbend', 'rbend'}:
            # We need to keep the rbarc parameter from the parent element.
            # If rbarc = True, then rbend length is the straight length.
            # If rbarc = False, then the length is the arc length, as for sbend.
            length = params.get('length', 0)

            if parent_name == 'rbend':
                rbarc = self._parameter_cache[name].get('rbarc', self.rbarc)
                if rbarc and 'length' in params:
                    params['length_straight'] = params.pop('length')

            # Default MAD-X behaviour is to take k0 from h only if k0 is not
            # given. We need to replicate this behaviour. Ideally we should
            # evaluate expressions here, but that's tricky.
            if self._parameter_cache[name].get('k0', 0) == 0:
                params['k0_from_h'] = True
            else:
                params['k0_from_h'] = False

            if (k2 := params.pop('k2', None)) and length:
                params['knl'] = [0, 0, k2 * length]
            if (k1s := params.pop('k1s', None)) and length:
                params['ksl'] = [0, k1s * length]
            if (hgap := params.pop('hgap', None)):
                params['edge_entry_hgap'] = hgap
                params['edge_exit_hgap'] = hgap

        elif parent_name in {'rfcavity', 'rfmultipole'}:
            if (lag := params.pop('lag', None)):
                params['lag'] = lag * 360
            if (volt := params.pop('volt', None)):
                params['voltage'] = volt * 1e6
            if (freq := params.pop('freq', None)):
                params['frequency'] = freq * 1e6
            if 'harmon' in params:
                # harmon * beam.beta * clight / sequence.length
                # raise NotImplementedError
                pass

        elif parent_name == 'multipole':
            if (knl := params.pop('knl', None)):
                params['knl'] = knl
            if (ksl := params.pop('ksl', None)):
                params['ksl'] = ksl
            if params.pop('lrad', None):
                _warn(f'Multipole `{name}` was specified with a length, ignoring!')

        elif parent_name == 'vkicker':
            if (kick := params.pop('kick', None)):
                params['ksl'] = [kick]

        elif parent_name == 'hkicker':
            if (kick := params.pop('kick', None)):
                params['knl'] = [-kick]

        elif parent_name in {'kicker', 'tkicker'}:
            if (vkick := params.pop('vkick', None)):
                params['ksl'] = [vkick]
            if (hkick := params.pop('hkick', None)):
                params['knl'] = [-hkick]

        if 'edge_entry_fint' in params and 'edge_exit_fint' not in params:
            params['edge_exit_fint'] = params['edge_entry_fint']
            # TODO: Technically MAD-X behaviour is that if edge_exit_fint < 0
            #  then we take edge_entry_fint as edge_exit_fint. But also,
            #  edge_entry_fint is the default value for edge_exit_fint.
            #  To implement this (unhinged?) feature faithfully we'd need to
            #  evaluate the expression here and ideally have a dynamic if-then
            #  expression... Instead, let's just pretend that edge_exit_fint
            #  should be taken as is, and hope no one relies on it being < 0.

        return params

    def _build_aperture(self, name, aper_name, params, force=False):
        """Build a Xtrack aperture for element `name` with  `params`.

        Parameters
        ----------
        name : str
            The name of the element for which to build the aperture.
        aper_name : str
            The name of the aperture element to be created.
        params : dict
            The parameters of the element, including the aperture parameters.
        force : bool, optional
            If True, the ``force`` parameter is passed to the builder when
            creating the aperture element: this will overwrite any existing
            aperture element with the same name. If False, an error will be
            raised if an aperture element with the same name already exists.

        Returns
        -------
        The name of the generated aperture element in the environment, or None.

        Notes:
        ------
        Currently supports all the basic MAD-X aperture types, however when
        ``aper_vx`` or ``aper_vy`` are given, the aperture is assumed to be
        simply a polygon, instead of applying the MAD-X logic (testing first for
        a simple shape and then for a polygon).
        """
        if not {'apertype', 'aperture', 'aper_vx', 'aper_vy'} & set(params):
            # No aperture parameters, nothing to do
            return

        if 'aper_vx' in params or 'aper_vy' in params:
            apertype = 'polygon'
            aperture = None
        else:
            apertype = params.pop('apertype', None) or self._mad_base_type(name)
            aperture = params.pop('aperture', None)

        if apertype not in _APERTURE_TYPES:
            raise ValueError(
                f'The aperture type for the element `{name}` (inferred to be '
                f'`{apertype}`) is not recognised.'
            )

        x_offset, y_offset = params.pop('aper_offset', (0, 0))
        if params.pop('aper_tol', None):
            _warn(f'Aperture tolerance (`{name}`) is not supported, ignoring.')
        aper_params = {
            'rot_s_rad': params.pop('aper_tilt', 0),
            'shift_x': x_offset,
            'shift_y': y_offset,
        }

        if apertype == 'circle':
            aper_params['a'] = aper_params['b'] = aperture[0]

        elif apertype == 'ellipse':
            aper_params['a'] = aperture[0]
            aper_params['b'] = aperture[1]

        elif apertype == 'rectangle':
            aper_params['min_x'] = -aperture[0]
            aper_params['max_x'] = aperture[0]
            aper_params['min_y'] = -aperture[1]
            aper_params['max_y'] = aperture[1]

        elif apertype == 'rectellipse':
            aper_params['max_x'] = aperture[0]
            aper_params['max_y'] = aperture[1]
            aper_params['a'] = aperture[2]
            aper_params['b'] = aperture[3]

        elif apertype == 'racetrack':
            aper_params['min_x'] = -aperture[0]
            aper_params['max_x'] = aperture[0]
            aper_params['min_y'] = -aperture[1]
            aper_params['max_y'] = aperture[1]
            aper_params['a'] = aperture[2]
            aper_params['b'] = aperture[3]

        elif apertype == 'octagon':
            # In MAD the octagon is defined with {w/2, h/2, phi_1, phi_2},
            # where w and h are respectively the width and height of the
            # rectangle that circumscribes the octagon, and phi_1 and phi_2
            # are the two angles sustaining the cut corner in the first
            # quadrant, given in radians, and with phi_1 < phi_2.
            half_w, half_h, phi_1, phi_2 = aperture
            y_right_corner = half_w * self.env.functions.tan(phi_1)
            x_top_corner = half_h / self.env.functions.tan(phi_2)
            top_x_vertices = [half_w, x_top_corner, -x_top_corner, -half_w]
            x_vertices = top_x_vertices + top_x_vertices[::-1]
            right_y_vertices = [y_right_corner, half_h, half_h, y_right_corner]
            y_vertices = right_y_vertices + [-y for y in right_y_vertices]
            aper_params['x_vertices'] = x_vertices
            aper_params['y_vertices'] = y_vertices

        elif apertype == 'polygon':
            if (aper_vx := params.pop('aper_vx', None)):
                aper_params['x_vertices'] = aper_vx
            if (aper_vy := params.pop('aper_vy', None)):
                aper_params['y_vertices'] = aper_vy

        return self.env.new(aper_name, _APERTURE_TYPES[apertype], force=force,
                            **aper_params)

    def _collect_hierarchy(self, parsed_dict: MadxOutputType):
        """Collect the base Madx types of all defined elements."""
        hierarchy = {}

        def _descend_into_line(line_params):
            if line_params['parent'] == 'line':
                return

            for name, elem_params in line_params['elements']:
                if (parent := elem_params.get('parent', None)):
                    hierarchy[name] = [parent] + hierarchy.get(parent, [])

                if parent == 'sequence':
                    _descend_into_line(elem_params)

        for name, el_params in parsed_dict['elements'].items():
            parent = el_params['parent']
            hierarchy[name] = [parent] + hierarchy.get(parent, [])

        for line_name, line_params in parsed_dict['lines'].items():
            _descend_into_line(line_params)

        return hierarchy


    def _mad_base_type(self, element_name: str):

        if element_name in self._madx_elem_hierarchy:
            return self._madx_elem_hierarchy[element_name][-1]

        if element_name not in self._builtin_types:
            raise ValueError(
                f'Cannot identify the MAD-X base type of element `{element_name}`!'
            )

        return element_name

    def _is_standalone_aperture(self, element_name: str):
        return self._mad_base_type(element_name) in _APERTURE_TYPES


def load_madx_lattice(file=None, string=None, reverse_lines=None):

    if file is not None and string is not None:
        raise ValueError('Only one of `file` or `string` can be provided!')

    if file is None and string is None:
        raise ValueError('Either `file` or `string` must be provided!')

    loader = MadxLoader()

    if file is not None:
        loader.load_file(file)
    elif string is not None:
        loader.load_string(string)
    else:
        raise ValueError('Something went wrong!')

    env = loader.env

    if reverse_lines:
        print('Reversing lines:', reverse_lines)
        rlines = {}
        for nn in reverse_lines:
            ll = env.lines[nn]
            llr = ll.copy()

            for enn in llr.element_names:
                _reverse_element(llr, enn)

            llr.discard_tracker()
            llr.element_names = llr.element_names[::-1]

            rlines[nn] = llr

        all_lines = {}
        for nn in env.lines.keys():
            if nn in rlines:
                all_lines[nn] = rlines[nn]
            else:
                all_lines[nn] = env.lines[nn]

        new_env = xt.Environment(lines=all_lines)

        # Adapt builders
        for nn in env.lines.keys():
            bb = env.lines[nn].builder.__class__(new_env)
            bb.__dict__.update(env.lines[nn].builder.__dict__)
            bb.env = new_env
            this_rename = new_env.lines[nn]._renamed_elements
            for cc in bb.components:
                cc.name = this_rename.get(cc.name, cc.name)
                cc.from_ = this_rename.get(cc.from_, cc.from_)

            if nn in reverse_lines:
                length = env.lines[nn].get_length()
                bb.components = bb.components[::-1]
                for cc in bb.components:
                    if cc.at is not None:
                        if isinstance(cc.at, str) or isinstance(cc.at, float):
                            if cc.from_ is not None:
                                cc.at = -cc.at
                            else:
                                cc.at = length - cc.at
            new_env.lines[nn].builder = bb

        # Add to new environment elements that were not in any line
        elems_not_in_lines = set(env.elements.keys())
        for nn in env.lines.keys():
            elems_not_in_lines -= set(env.lines[nn].element_names)
        dummy_line = env.new_line(components=list(elems_not_in_lines))
        new_env.import_line(line=dummy_line, line_name='__DUMMY__')
        del new_env.lines['__DUMMY__'] # keep the elements but not the line

        env = new_env
        print('Done reversing lines')

    return env


import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=450e9, mass0=xt.PROTON_MASS_EV, q0=1)
env.vars.default_to_zero = True

env.new('mq', 'Quadrupole', length=0.8)
env.new('bpm', 'Marker')
env.new('corrector', 'Multipole', knl=[0])

line = env.new_line(components=[

    env.new('line.start', 'Marker'),
    env.new('line.end', 'Marker', at=12.),

    env.new('mq1', 'mq', k1='kq1', at=3.),
    env.new('mq2', 'mq', k1='kq2', at=5.),
    env.new('mq3', 'mq', k1='kq3', at=7.),
    env.new('mq4', 'mq', k1='kq4', at=9.),

    env.new('bpm1', 'bpm', at='mq1@start'),
    env.new('bpm2', 'bpm', at='mq2@start'),
    env.new('bpm3', 'bpm', at='mq3@start'),
    env.new('bpm4', 'bpm', at='mq4@start'),

    env.new('corrector1', 'corrector', at='mq1@start'),
    env.new('corrector2', 'corrector', at='mq2@start'),
    env.new('corrector3', 'corrector', at='mq3@start'),
    env.new('corrector4', 'corrector', at='mq4@start'),

    env.new('bumper1', 'corrector', at=1., knl=['k0l_bumper1'], ksl=['k0sl_bumper1']),
    env.new('bumper2', 'corrector', at=2., knl=['k0l_bumper2'], ksl=['k0sl_bumper2']),
])

env.set(['mq1', 'mq2', 'mq3', 'mq4'], shift_x=1e-3, shift_y=1.5e-3,
        rot_s_rad=np.deg2rad(30.))

# Steer to enter at the center of the first quad
line.match(
    betx=1., bety=1.,
    vary=xt.VaryList(['k0l_bumper1', 'k0l_bumper2', 'k0sl_bumper1', 'k0sl_bumper2'],
                     step=1e-6),
    targets=xt.TargetSet(x=1e-3, px=0, y=1.5e-3, py=0, at='bpm1'),
)

tt = line.get_table()
tt_quad = tt.rows['mq.*']

env['kq1'] = 0.02
env['kq2'] = -0.02
env['kq3'] = 0.02
env['kq4'] = -0.02

dx = 2e-3

bpm_alignment ={
    'bpm1': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
    'bpm2': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
    'bpm3': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
    'bpm4': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
}

tw0 = line.twiss(betx=100, bety=80)

line.steering_monitors_x = ['bpm1', 'bpm2', 'bpm3', 'bpm4']
line.steering_monitors_y = ['bpm1', 'bpm2', 'bpm3', 'bpm4']
line.steering_correctors_x = ['corrector1', 'corrector3', ]
line.steering_correctors_y = ['corrector2', 'corrector4']

tw = line.twiss(betx=1, bety=1, x=1e-3+dx*np.cos(np.deg2rad(30)), y=1.5e-3+dx*np.sin(np.deg2rad(30)))


correction = line.correct_trajectory(twiss_table=tw0,
                                     x_init=dx * np.cos(np.deg2rad(30)),
                                     y_init=dx * np.sin(np.deg2rad(30)),
                                     start='line.start', end='line.end',
                                     monitor_alignment=bpm_alignment,
                                     run=False)

correction.correct(n_iter=1)

# Check that there is no vertical reading in the tilted bpm BPMs
xo.assert_allclose(correction.y_correction._position_before, 0, atol=1e-15, rtol=0)



import json
import pathlib

import xtrack as xt
from cpymad.madx import Madx
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(__file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_multiline_metadata(test_context):
    with open(test_data_folder / 'hllhc14_no_errors_with_coupling_knobs/line_b1.json', 'r') as fid:
        dct_b1 = json.load(fid)
        input_line = xt.Line.from_dict(dct_b1)

    mad = Madx(stdout=False)
    mad.call(str(test_data_folder / 'hllhc14_no_errors_with_coupling_knobs/lhcb1_seq.madx'))
    mad.use(sequence='lhcb1')

    input_line_co_ref = xt.Line.from_madx_sequence(
        mad.sequence.lhcb1,
        deferred_expressions=True,
        expressions_for_element_types=('kicker', 'hkicker', 'vkicker'),
    )

    # Build the collider with the default constructor
    collider = xt.Environment(
        lines={'lhcb1': input_line.copy(), 'lhcb1_co_ref': input_line_co_ref.copy()}
    )
    collider['lhcb1_co_ref'].particle_ref = collider['lhcb1'].particle_ref.copy()

    # Test the dump and load into/from dictionnary without metadata
    collider = xt.Environment.from_dict(collider.to_dict())

    # Add metadata
    collider.metadata = {
        'config_knobs_and_tuning': {
            'knob_settings': {
                'on_x1': 135.0,
            },
        },
        'qx': {'lhcb1': 62.31, 'lhcb2': 62.31},
        'delta_cmr': 0.0,
    }

    # Test the dump and load into/from dictionnary with metadata
    collider_copy = xt.Environment.from_dict(collider.to_dict())
    
    # Assert that both metadata are still identical
    assert collider.metadata == collider_copy.metadata
    
    # Mutate the copy and check that the original is not changed
    collider_copy.metadata['qx']['lhcb1'] = collider.metadata['qx']['lhcb1'] + 1
    assert collider.metadata != collider_copy.metadata
    
    # Ensuire trackers can still be built
    collider.build_trackers(_context=test_context)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy.constants import e as qe
from scipy.constants import c as clight
from scipy.constants import epsilon_0

import xtrack as xt
import xobjects as xo

context = xo.context_default

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)

for factor in [0.2, 0.5, 1, 2, 5, 10, 20, 100]:
    theta_bend = 0.05 * factor
    L_bend = 5. * factor

    dipole_ave = xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend,
                              radiation_flag=1, _context=context)
    dipole_rnd = xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend,
                              radiation_flag=2, _context=context)

    particles_ave = xt.Particles(
            _context=context,
            p0c=5e9, # 5 GeV
            x=np.zeros(1000000),
            px=1e-4,
            py=-1e-4,
            mass0=xt.ELECTRON_MASS_EV)
    particles_rnd = particles_ave.copy()

    dct_ave_before = particles_ave.to_dict()
    dct_rng_before = particles_rnd.to_dict()

    particles_ave._init_random_number_generator()
    particles_rnd._init_random_number_generator()

    dipole_ave.track(particles_ave)
    dipole_rnd.track(particles_rnd)

    dct_ave = particles_ave.to_dict()
    dct_rng = particles_rnd.to_dict()

    vals, bins = np.histogram((dct_rng['delta']-dct_ave['delta']), bins=100,
                              range=(-2e-5, 0))

    plt.semilogy(bins[:-1], vals)


plt.show()



import numpy as np
from cpymad.madx import Madx
import xtrack as xt

from xtrack.slicing import Teapot, Strategy

# hllhc15 can be found at git@github.com:lhcopt/hllhc15.git

thin = False
kill_fringes_and_edges = True

mq_slice_list = np.arange(2, 12, 2)

diff_qx_xsuite_list = []
diff_qy_xsuite_list = []
diff_qx_mad_list = []
diff_qy_mad_list = []


for ii, mq_sl in enumerate(mq_slice_list):
    mad = Madx()

    mad.input(f"""
    call,file="../../../hllhc15/util/lhc.seq";
    call,file="../../../hllhc15/hllhc_sequence.madx";
    call,file="../../../hllhc15/toolkit/macro.madx";
    seqedit,sequence=lhcb1;flatten;cycle,start=IP7;flatten;endedit;
    seqedit,sequence=lhcb2;flatten;cycle,start=IP7;flatten;endedit;
    exec,mk_beam(7000);
    call,file="../../../hllhc15/round/opt_round_150_1500.madx";
    {('exec,myslice;' if thin else '')}
    exec,check_ip(b1);
    exec,check_ip(b2);
    // acbv11.r8b1 = 2e-6;
    """)

    mad.use(sequence="lhcb1")
    seq = mad.sequence.lhcb1
    mad.twiss()

    if ii == 0:
        line0 = xt.Line.from_madx_sequence(
            mad.sequence.lhcb1, allow_thick=True, deferred_expressions=True)
        line0.particle_ref = xt.Particles(mass0=seq.beam.mass*1e9, gamma0=seq.beam.gamma)
        line0.twiss_default['method'] = '4d'
        line0.twiss_default['matrix_stability_tol'] = 100
        line0.build_tracker()
        tw0_before = line0.twiss()

    line = line0.copy()

    n_slice_bends = 4
    n_slice_quads = 20
    n_slice_mb = 2
    # n_slice_mq = 10
    n_slice_mq = mq_sl
    n_slice_mqt = 5
    n_slice_mqx = 60

    slicing_strategies = [
        Strategy(slicing=Teapot(1)),  # Default catch-all as in MAD-X
        Strategy(slicing=None, element_type=xt.Solenoid),
        Strategy(slicing=Teapot(n_slice_bends), element_type=xt.Bend),
        Strategy(slicing=Teapot(n_slice_quads), element_type=xt.Quadrupole),
        Strategy(slicing=Teapot(n_slice_mb), name=r'^mb\..*'),
        Strategy(slicing=Teapot(n_slice_mq), name=r'^mq\..*'),
        Strategy(slicing=Teapot(n_slice_mq), name=r'^mqt.*'),
        Strategy(slicing=Teapot(n_slice_mqx), name=r'^mqx.*'),
    ]

    print("Slicing thick elements...")
    line.slice_thick_elements(slicing_strategies)

    print("Building tracker...")
    line.build_tracker()

    tw0_after = line0.twiss()
    tw_after = line.twiss()

    # Compare tunes
    print("Tunes before slicing:")
    print(f"Thick: qx = {tw0_before.qx} \tqy = {tw0_before.qy}")
    print(f"Thin:  qx = {tw_after.qx} \tqy = {tw_after.qy}")
    print(f"Diffs: qx = {tw_after.qx - tw0_before.qx} \tqy = {tw_after.qy - tw0_before.qy}")

    thick_arc_bends = False
    thick_arc_quads = False

    twmad_thick = mad.twiss().dframe()

    mad.input(
    f'''
    select, flag=makethin, clear;
    select, flag=makethin, class=mb, slice={('0' if thick_arc_bends else n_slice_mb)};
    select, flag=makethin, class=mq, slice={('0' if thick_arc_quads else n_slice_mq)};
    select, flag=makethin, class=mqxa,       slice={n_slice_mqx};  !old triplet
    select, flag=makethin, class=mqxb,       slice={n_slice_mqx};  !old triplet
    select, flag=makethin, class=mqxc,       slice={n_slice_mqx};  !new mqxa (q1,q3)
    select, flag=makethin, class=mqxd,       slice={n_slice_mqx};  !new mqxb (q2a,q2b)
    select, flag=makethin, class=mqxfa,      slice={n_slice_mqx};  !new (q1,q3 v1.1)
    select, flag=makethin, class=mqxfb,      slice={n_slice_mqx};  !new (q2a,q2b v1.1)
    select, flag=makethin, class=mbxa,       slice={n_slice_bends};   !new d1
    select, flag=makethin, class=mbxf,       slice={n_slice_bends};   !new d1 (v1.1)
    select, flag=makethin, class=mbrd,       slice={n_slice_bends};   !new d2 (if needed)
    select, flag=makethin, class=mqyy,       slice={n_slice_mq};   !new q4
    select, flag=makethin, class=mqyl,       slice={n_slice_mq};   !new q5
    select, flag=makethin, pattern=mbx\.,    slice={n_slice_bends};
    select, flag=makethin, pattern=mbrb\.,   slice={n_slice_bends};
    select, flag=makethin, pattern=mbrc\.,   slice={n_slice_bends};
    select, flag=makethin, pattern=mbrs\.,   slice={n_slice_bends};
    select, flag=makethin, pattern=mbh\.,    slice={n_slice_bends};
    select, flag=makethin, pattern=mqwa\.,   slice={n_slice_quads};
    select, flag=makethin, pattern=mqwb\.,   slice={n_slice_quads};
    select, flag=makethin, pattern=mqy\.,    slice={n_slice_quads};
    select, flag=makethin, pattern=mqm\.,    slice={n_slice_quads};
    select, flag=makethin, pattern=mqmc\.,   slice={n_slice_quads};
    select, flag=makethin, pattern=mqml\.,   slice={n_slice_quads};
    select, flag=makethin, pattern=mqtlh\.,  slice={n_slice_mqt};
    select, flag=makethin, pattern=mqtli\.,  slice={n_slice_mqt};
    select, flag=makethin, pattern=mqt\.  ,  slice={n_slice_mqt};

    beam;
    use,sequence=lhcb1; makethin,sequence=lhcb1,makedipedge=true,style=teapot;
    use,sequence=lhcb2; makethin,sequence=lhcb2,makedipedge=true,style=teapot;
    use, sequence=lhcb2;
    use, sequence=lhcb1;

    ''')

    twmad_thin = mad.twiss().dframe()

    # Compare tunes
    print(f"Xsuite thick: qx = {tw0_before.qx} \tqy = {tw0_before.qy}")
    print(f"Xsuite thin:  qx = {tw_after.qx} \tqy = {tw_after.qy}")

    qx_mad_before = twmad_thick.mux[-1]
    qy_mad_before = twmad_thick.muy[-1]
    qx_mad_after = twmad_thin.mux[-1]
    qy_mad_after = twmad_thin.muy[-1]

    print(f"MAD-X thick: qx = {qx_mad_before} \tqy = {qy_mad_before}")
    print(f"MAD-X thin:  qx = {qx_mad_after} \tqy = {qy_mad_after}")

    print(f"Xsuite diffs: qx = {tw_after.qx - tw0_before.qx} \tqy = {tw_after.qy - tw0_before.qy}")
    print(f"MAD-X diffs: qx = {qx_mad_after - qx_mad_before} \tqy = {qy_mad_after - qy_mad_before}")

    diff_qx_xsuite_list.append(tw_after.qx - tw0_before.qx)
    diff_qy_xsuite_list.append(tw_after.qy - tw0_before.qy)
    diff_qx_mad_list.append(qx_mad_after - qx_mad_before)
    diff_qy_mad_list.append(qy_mad_after - qy_mad_before)

import matplotlib.pyplot as plt
plt.close('all')
plt.plot(mq_slice_list, diff_qx_xsuite_list, label='qx xsuite', color='g')
plt.plot(mq_slice_list, diff_qy_xsuite_list, label='qy xsuite', color='m')
plt.plot(mq_slice_list, diff_qx_mad_list, '--', label='qx mad', color='b')
plt.plot(mq_slice_list, diff_qy_mad_list, '--', label='qy mad', color='r')
plt.legend()
plt.xlabel('n_slice_mq')
plt.ylabel(r'$Q_{thin} - Q_{thick}$')
plt.grid()
plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from .masses import PROTON_MASS_EV, ELECTRON_MASS_EV, MUON_MASS_EV, Pb208_MASS_EV
from .particles import Particles, reference_from_pdg_id, LAST_INVALID_STATE


def enable_pyheadtail_interface():
    import xpart.pyheadtail_interface.pyhtxtparticles as pp
    import xpart as xp
    import xtrack as xt
    xp.Particles = pp.PyHtXtParticles
    xt.Particles = pp.PyHtXtParticles


def disable_pyheadtail_interface():
    import xpart as xp
    import xtrack as xt
    xp.Particles = Particles
    xt.Particles = Particles



import json

import xtrack as xt

# import a line and add reference particle
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7e12)

# Build the tracker
line.build_tracker()

# Track some particles with frozen longitudinal coordinates
particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
line.track(particles, num_turns=10, freeze_longitudinal=True)
print(particles.delta) # gives [0.001 0.001 0.001], same as initial value

# Twiss with frozen longitudinal coordinates (needs to be 4d)
twiss = line.twiss(method='4d', freeze_longitudinal=True)
print(twiss.slip_factor) # gives 0 (no longitudinal motion)

# Track some particles with unfrozen longitudinal coordinates
particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
line.track(particles, num_turns=10)
print(particles.delta) # gives [0.00099218, ...], different from initial value

# Twiss with unfrozen longitudinal coordinates (can be 6d)
twiss = line.twiss(method='6d')
print(twiss.slip_factor) # gives 0.00032151, from longitudinal motion



import numpy as np
import xtrack as xt
import xobjects as xo
from scipy.constants import c as clight
from scipy.constants import e as qe

from cpymad.madx import Madx

fname = 'fccee_z'; pc_gev = 45.6
# fname = 'fccee_t'; pc_gev = 182.5

# Load from MAD-X
mad = Madx()
mad.call('../../test_data/fcc_ee/' + fname + '.seq')
mad.beam(particle='positron', pc=pc_gev)
mad.use('fccee_p_ring')

line = xt.Line.from_madx_sequence(mad.sequence.fccee_p_ring, allow_thick=True,
                                  deferred_expressions=True)
line.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV,
                                 gamma0=mad.sequence.fccee_p_ring.beam.gamma)
line.cycle('ip.4', inplace=True)
line.append_element(element=xt.Marker(), name='ip.4.l')
line.build_tracker()

# Get table with the elements
tt = line.get_table()

# Save voltage values
line.vars['voltca1_ref'] = line.vv['voltca1']
if 'voltca2' in line.vars.keys():
    line.vars['voltca2_ref'] = line.vv['voltca2']
else:
    line.vars['voltca2_ref'] = 0

# Switch off the cavities
line.vars['voltca1'] = 0
line.vars['voltca2'] = 0

# Load solenoid Bz data
bz_data_file = '../../test_data/fcc_ee/Bz_closed_before_quads.dat'
import pandas as pd
bz_df = pd.read_csv(bz_data_file, sep='\s+', skiprows=1, names=['z', 'Bz'])

# Solenoid parameters
theta_tilt = 15e-3 # rad
l_beam = 4.4       # Length of the solenoid along the beam trajectory
l_solenoid = l_beam * np.cos(theta_tilt)
ds_sol_start = -l_beam / 2 # from ip
ds_sol_end = +l_beam / 2   # from ip
ip_sol = 'ip.1'

# Prepare solenoid slices
s_sol_slices = np.linspace(-l_solenoid/2, l_solenoid/2, 1001)
bz_sol_slices = np.interp(s_sol_slices, bz_df.z, bz_df.Bz)
bz_sol_slices[0] = 0
bz_sol_slices[-1] = 0

# Compute strength of solenoid slices
P0_J = line.particle_ref.p0c[0] * qe / clight
brho = P0_J / qe / line.particle_ref.q0
ks = 0.5 * (bz_sol_slices[:-1] + bz_sol_slices[1:]) / brho

# Length and entry position of solenoid slices
l_sol_slices = np.diff(s_sol_slices)
s_sol_slices_entry = s_sol_slices[:-1]

# Build the slices
sol_slices = []
for ii in range(len(s_sol_slices_entry)):
    sol_slices.append(xt.Solenoid(length=l_sol_slices[ii], ks=0)) # Off for now

s_ip = tt['s', ip_sol]

# Install markers at solenoid entry and exit
line.discard_tracker()
line.insert('sol_start_'+ip_sol, xt.Marker(), at=s_ip + ds_sol_start)
line.insert('sol_end_'+ip_sol, xt.Marker(), at=s_ip + ds_sol_end)

# Build reference frame transformations at ebtry and exit of solenoid
sol_start_tilt = xt.YRotation(angle=-theta_tilt * 180 / np.pi)
sol_end_tilt = xt.YRotation(angle=+theta_tilt * 180 / np.pi)
sol_start_shift = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt))
sol_end_shift = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt))

line.element_dict['sol_start_tilt_'+ip_sol] = sol_start_tilt
line.element_dict['sol_end_tilt_'+ip_sol] = sol_end_tilt
line.element_dict['sol_start_shift_'+ip_sol] = sol_start_shift
line.element_dict['sol_end_shift_'+ip_sol] = sol_end_shift

line.element_dict['sol_entry_'+ip_sol] = xt.Solenoid(length=0, ks=0)
line.element_dict['sol_exit_'+ip_sol] = xt.Solenoid(length=0, ks=0)
line.element_dict['sol_zeta_shift_'+ip_sol] = xt.ZetaShift(dzeta=-(l_beam - l_solenoid))

# Add slices to the elements pot
sol_slice_names = []
sol_slice_names.append('sol_entry_'+ip_sol)
for ii in range(len(s_sol_slices_entry)):
    nn = f'sol_slice_{ii}_{ip_sol}'
    line.element_dict[nn] = sol_slices[ii]
    sol_slice_names.append(nn)
sol_slice_names.append('sol_exit_'+ip_sol)

# Insert new elements in the names list
tt = line.get_table()
names_upstream = list(tt.rows[:'sol_start_'+ip_sol].name)
names_downstream = list(tt.rows['sol_end_'+ip_sol:].name[:-1]) # -1 to exclude '_end_point' added by the table
element_names = (names_upstream
                 + ['sol_start_tilt_'+ip_sol, 'sol_start_shift_'+ip_sol]
                 + sol_slice_names
                 + ['sol_end_shift_'+ip_sol, 'sol_end_tilt_'+ip_sol]
                 + ['sol_zeta_shift_'+ip_sol]
                 + names_downstream)
line.element_names = element_names

# re-insert the ip
line.element_dict.pop(ip_sol)
tt = line.get_table()
line.insert(ip_sol, xt.Marker(),
        at = 0.5 * (tt['s', 'sol_start_'+ip_sol] + tt['s', 'sol_end_'+ip_sol]))

line.build_tracker()

# Set strength
line.vars['on_sol_'+ip_sol] = 0
for ii in range(len(s_sol_slices_entry)):
    nn = f'sol_slice_{ii}_{ip_sol}'
    line.element_refs[nn].ks = ks[ii] * line.vars['on_sol_'+ip_sol]

tt = line.get_table()

# Inspect
tt.rows['sol_start_ip.1':'sol_end_ip.1'].show()

# Add skew quadrupole knobs to the final focus quadrupoles
line.vars['on_corr_ip.1'] = 1
line.vars['ks0.r1'] = 0
line.vars['ks1.r1'] = 0
line.vars['ks2.r1'] = 0
line.vars['ks3.r1'] = 0
line.vars['ks4.r1'] = 0
line.vars['ks0.l1'] = 0
line.vars['ks1.l1'] = 0
line.vars['ks2.l1'] = 0
line.vars['ks3.l1'] = 0
line.vars['ks4.l1'] = 0

line.element_refs['qc1r1.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks0.r1']
line.element_refs['qc2r1.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks1.r1']
line.element_refs['qc2r2.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks2.r1']
line.element_refs['qc1r2.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks3.r1']
line.element_refs['qc1l1.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks0.l1']
line.element_refs['qc2l1.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks1.l1']
line.element_refs['qc2l2.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks2.l1']
line.element_refs['qc1l2.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks3.l1']

# Add correction knobs normal strength of the final focus quadrupoles
line.vars['corr_k0.r1'] = 0
line.vars['corr_k1.r1'] = 0
line.vars['corr_k2.r1'] = 0
line.vars['corr_k3.r1'] = 0
line.vars['corr_k4.r1'] = 0
line.vars['corr_k0.l1'] = 0
line.vars['corr_k1.l1'] = 0
line.vars['corr_k2.l1'] = 0
line.vars['corr_k3.l1'] = 0
line.vars['corr_k4.l1'] = 0

line.element_refs['qc1r1.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k0.r1']
line.element_refs['qc2r1.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k1.r1']
line.element_refs['qc2r2.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k2.r1']
line.element_refs['qc1r2.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k3.r1']
line.element_refs['qc1l1.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k0.l1']
line.element_refs['qc2l1.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k1.l1']
line.element_refs['qc2l2.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k2.l1']
line.element_refs['qc1l2.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k3.l1']

tw_thick_no_rad = line.twiss(method='4d')

# Slice the lattice
Strategy = xt.Strategy
Teapot = xt.Teapot
slicing_strategies = [
    Strategy(slicing=None),  # Default catch-all as in MAD-X
    Strategy(slicing=Teapot(3), element_type=xt.Bend),
    Strategy(slicing=Teapot(2), element_type=xt.Sextupole),
    # Strategy(slicing=Teapot(50), element_type=xt.Quadrupole), # Starting point
    Strategy(slicing=Teapot(5), name=r'^qf.*'),
    Strategy(slicing=Teapot(5), name=r'^qd.*'),
    Strategy(slicing=Teapot(5), name=r'^qfg.*'),
    Strategy(slicing=Teapot(5), name=r'^qdg.*'),
    Strategy(slicing=Teapot(5), name=r'^ql.*'),
    Strategy(slicing=Teapot(5), name=r'^qs.*'),
    Strategy(slicing=Teapot(10), name=r'^qb.*'),
    Strategy(slicing=Teapot(10), name=r'^qg.*'),
    Strategy(slicing=Teapot(10), name=r'^qh.*'),
    Strategy(slicing=Teapot(10), name=r'^qi.*'),
    Strategy(slicing=Teapot(10), name=r'^qr.*'),
    Strategy(slicing=Teapot(10), name=r'^qu.*'),
    Strategy(slicing=Teapot(10), name=r'^qy.*'),
    Strategy(slicing=Teapot(50), name=r'^qa.*'),
    Strategy(slicing=Teapot(50), name=r'^qc.*'),
    Strategy(slicing=Teapot(20), name=r'^sy\..*'),
    Strategy(slicing=Teapot(30), name=r'^mwi\..*'),
]
line.discard_tracker()
line.slice_thick_elements(slicing_strategies=slicing_strategies)

# Add dipole correctors
line.insert('mcb1.r1', xt.Multipole(knl=[0]), at='qc1r1.1_exit@start')
line.insert('mcb2.r1', xt.Multipole(knl=[0]), at='qc1r2.1_exit@start')
line.insert('mcb1.l1', xt.Multipole(knl=[0]), at='qc1l1.4_entry@start')
line.insert('mcb2.l1', xt.Multipole(knl=[0]), at='qc1l2.4_entry@start')

line.vars['acb1h.r1'] = 0
line.vars['acb1v.r1'] = 0
line.vars['acb2h.r1'] = 0
line.vars['acb2v.r1'] = 0
line.vars['acb1h.l1'] = 0
line.vars['acb1v.l1'] = 0
line.vars['acb2h.l1'] = 0
line.vars['acb2v.l1'] = 0

line.element_refs['mcb1.r1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb1h.r1']
line.element_refs['mcb2.r1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb2h.r1']
line.element_refs['mcb1.r1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb1v.r1']
line.element_refs['mcb2.r1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb2v.r1']
line.element_refs['mcb1.l1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb1h.l1']
line.element_refs['mcb2.l1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb2h.l1']
line.element_refs['mcb1.l1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb1v.l1']
line.element_refs['mcb2.l1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb2v.l1']

assert line.element_names[-1] == 'ip.4.l'
assert line.element_names[0] == 'ip.4'

opt = line.match(
    method='4d',
    start='ip.4', end='ip.4.l',
    init=tw_thick_no_rad,
    vary=xt.VaryList(['k1qf4', 'k1qf2', 'k1qd3', 'k1qd1',], step=1e-8,
    ),
    targets=[
        xt.TargetSet(at=xt.END, mux=tw_thick_no_rad.qx, muy=tw_thick_no_rad.qy, tol=1e-5),
    ]
)
opt.solve()
tw_thin_no_rad = line.twiss(method='4d')

# Check partition numbers
line.to_json(fname + '_with_sol.json')

line.vars['voltca1'] = line.vars['voltca1_ref']
line.vars['voltca2'] = line.vars['voltca2_ref']

line.build_tracker()

line.configure_radiation(model='mean')
line.compensate_radiation_energy_loss()

two = line.twiss(eneloss_and_damping=True, particle_on_co=line.particle_ref.copy())
tw = line.twiss(eneloss_and_damping=True)

print(tw.partition_numbers)